"""
.. module:: cobaya.grid_tools.batchjob

:Synopsis: Classes for jobs and datasets
:Author: Antony Lewis, modified by Jesus Torrado

"""

# TODO many things not yet updated for cobaya formats

import os
import shutil
import pickle
import copy
import sys
import time
from typing import Callable
from getdist import types, IniFile
from getdist.mcsamples import loadMCSamples

from .conventions import input_folder, script_folder, log_folder


def grid_cache_file(directory):
    return os.path.abspath(directory) + os.sep + 'batch.pyobj'


def resetGrid(directory):
    fname = grid_cache_file(directory)
    if os.path.exists(fname):
        os.remove(fname)


def readobject(directory=None):
    # load this here to prevent circular
    from cobaya.grid_tools import gridconfig

    if directory is None:
        directory = sys.argv[1]
    fname = grid_cache_file(directory)
    if not os.path.exists(fname):
        if gridconfig.pathIsGrid(directory):
            return gridconfig.makeGrid(directory, read_only=True, interactive=False)
        return None
    try:
        config_dir = os.path.abspath(directory) + os.sep + 'config'
        if os.path.exists(config_dir):
            # set path in case using functions defined
            # and hence imported from in settings file
            sys.path.insert(0, config_dir)
        with open(fname, 'rb') as inp:
            grid = pickle.load(inp)
        if not os.path.exists(grid.basePath):
            raise FileNotFoundError('Directory not found %s' % grid.basePath)
        return grid
    except Exception as e:
        print('Error loading cached batch object: ', e)
        resetGrid(directory)
        if gridconfig.pathIsGrid(directory):
            return gridconfig.makeGrid(directory, read_only=True, interactive=False)
        raise


def saveobject(obj, filename):
    with open(filename, 'wb') as output:
        pickle.dump(obj, output, pickle.HIGHEST_PROTOCOL)


def makePath(s):
    if not os.path.exists(s):
        os.makedirs(s)


def nonEmptyFile(fname):
    return os.path.exists(fname) and os.path.getsize(fname) > 0


def getCodeRootPath():
    return os.path.normpath(os.path.join(os.path.dirname(__file__), '..', '..')) + os.sep


class PropertiesItem:
    propertiesIniFile: Callable

    def propertiesIni(self):
        if os.path.exists(self.propertiesIniFile()):
            return IniFile(self.propertiesIniFile())
        else:
            ini = IniFile()
            ini.original_filename = self.propertiesIniFile()
            return ini


class DataSet:
    importanceNames: list
    importanceParams: list

    def __init__(self, names, params=None, covmat=None, dist_settings=None):
        if not dist_settings:
            dist_settings = {}
        if isinstance(names, str):
            names = [names]
        if params is None:
            params = [(name + '.ini') for name in names]
        else:
            params = self.standardizeParams(params)
        if covmat is not None:
            self.covmat = covmat
        self.names = names
        self.params = params
        # can be an array of items, either ini file name or dictionaries of parameters
        self.tag = "_".join(self.names)
        self.dist_settings = dist_settings

    def add(self, name, params=None, overrideExisting=True, dist_settings=None):
        if params is None:
            params = [name]
        params = self.standardizeParams(params)
        if dist_settings:
            self.dist_settings.update(dist_settings)
        if overrideExisting:
            self.params = params + self.params
            # can be an array of items, either ini file name or dictionaries of parameters
        else:
            self.params += params
        if name is not None:
            self.names += [name]
            self.tag = "_".join(self.names)
        return self

    def addEnd(self, name, params, dist_settings=None):
        if not dist_settings:
            dist_settings = {}
        return self.add(name, params, overrideExisting=False, dist_settings=dist_settings)

    def copy(self):
        return copy.deepcopy(self)

    def extendForImportance(self, names, params):
        data = copy.deepcopy(self)
        if '_post_' not in data.tag:
            data.tag += '_post_' + "_".join(names)
        else:
            data.tag += '_' + "_".join(names)
        data.importanceNames = names
        data.importanceParams = data.standardizeParams(params)
        data.names += data.importanceNames
        data.params += data.importanceParams
        return data

    def standardizeParams(self, params):
        if isinstance(params, dict) or isinstance(params, str):
            params = [params]
        for i in range(len(params)):
            if isinstance(params[i], str) and '.ini' not in params[i]:
                params[i] += '.ini'
        return params

    def hasName(self, name):
        if isinstance(name, str):
            return name in self.names
        else:
            return any(True for i in name if i in self.names)

    def hasAll(self, name):
        if isinstance(name, str):
            return name in self.names
        else:
            return all((i in self.names) for i in name)

    def tagReplacing(self, x, y):
        items = []
        for name in self.names:
            if name == x:
                if y != '':
                    items.append(y)
            else:
                items.append(name)
        return "_".join(items)

    def namesReplacing(self, dic):
        if dic is None:
            return self.names
        items = []
        for name in self.names:
            if name in dic:
                val = dic[name]
                if val:
                    items.append(val)
            else:
                items.append(name)
        return items

    def makeNormedDatatag(self, dic):
        return "_".join(sorted(self.namesReplacing(dic)))


class JobGroup:
    def __init__(self, name, params=None, importanceRuns=None, datasets=None):
        if importanceRuns is None:
            importanceRuns = []
        if params is None:
            params = [[]]
        if datasets is None:
            datasets = []
            self.params = params
            self.groupName = name
            self.importanceRuns = importanceRuns
            self.datasets = datasets


class ImportanceSetting:
    def __init__(self, names, inis=None, dist_settings=None, minimize=True):
        if not inis:
            inis = []
        self.names = names
        self.inis = inis
        self.dist_settings = dist_settings or {}
        self.want_minimize = minimize

    def wantImportance(self, _jobItem):
        return True


class ImportanceFilter(ImportanceSetting):
    # class for trivial importance sampling filters that can be done in python,
    # e.g. restricting a parameter to a new range

    def __init__(self, names, dist_settings=None, minimize=False):
        super().__init__(names, inis=[self], dist_settings=dist_settings,
                         minimize=minimize)


class JobItem(PropertiesItem):
    importanceTag: str
    importanceSettings: list
    importanceFilter: ImportanceFilter

    def __init__(self, path, param_set, data_set, base='base', minimize=True):
        self.param_set = param_set
        if not isinstance(data_set, DataSet):
            data_set = DataSet(data_set)
        self.data_set = data_set
        self.base = base
        self.paramtag = base + "_" + param_set
        self.datatag = data_set.tag
        self.name = self.paramtag + '_' + self.datatag
        self.batchPath = path
        self.relativePath = self.paramtag + os.sep + self.datatag + os.sep
        self.chainPath = path + self.relativePath
        self.chainRoot = self.chainPath + self.name
        self.distPath = self.chainPath + 'dist' + os.sep
        self.distRoot = self.distPath + self.name
        self.isImportanceJob = False
        self.importanceItems = []
        self.want_minimize = minimize
        self.result_converge = None
        self.group = None
        self.parent = None
        self.dist_settings = copy.copy(data_set.dist_settings)
        self.makeIDs()
        self.iniFile_path = input_folder
        self.iniFile_ext = ".yaml"
        self.scriptFile_path = script_folder
        self.logFile_path = log_folder

    def iniFile(self, variant=''):
        if not self.isImportanceJob:
            return (self.batchPath + self.iniFile_path + os.sep +
                    self.name + variant + self.iniFile_ext)
        else:
            return (self.batchPath + 'postIniFiles' + os.sep +
                    self.name + variant + self.iniFile_ext)

    def propertiesIniFile(self):
        return self.chainRoot + '.properties.ini'

    def isBurnRemoved(self):
        return self.propertiesIni().bool('burn_removed')

    def makeImportance(self, importanceRuns):
        for impRun in importanceRuns:
            if isinstance(impRun, ImportanceSetting):
                if not impRun.wantImportance(self):
                    continue
            else:
                if len(impRun) > 2 and not impRun[2].wantImportance(self):
                    continue
                impRun = ImportanceSetting(impRun[0], impRun[1])
            if len(set(impRun.names).intersection(self.data_set.names)) > 0:
                print('importance job duplicating parent data set: %s with %s' % (
                    self.name, impRun.names))
                continue
            data = self.data_set.extendForImportance(impRun.names, impRun.inis)
            job = JobItem(self.batchPath, self.param_set, data,
                          minimize=impRun.want_minimize)
            job.importanceTag = "_".join(impRun.names)
            job.importanceSettings = impRun.inis
            if '_post_' not in self.name:
                tag = '_post_' + job.importanceTag
            else:
                tag = '_' + job.importanceTag
            job.name = self.name + tag
            job.chainRoot = self.chainRoot + tag
            job.distPath = self.distPath
            job.chainPath = self.chainPath
            job.relativePath = self.relativePath
            job.distRoot = self.distRoot + tag
            job.datatag = self.datatag + tag
            job.isImportanceJob = True
            job.parent = self
            job.group = self.group
            job.dist_settings.update(impRun.dist_settings)
            if isinstance(impRun, ImportanceFilter):
                job.importanceFilter = impRun
            job.makeIDs()
            self.importanceItems.append(job)

    def makeNormedName(self, dataSubs=None):
        normed_params = "_".join(sorted(self.param_set))
        normed_data = self.data_set.makeNormedDatatag(dataSubs)
        normed_name = self.base
        if len(normed_params) > 0:
            normed_name += '_' + normed_params
        normed_name += '_' + normed_data
        return normed_name, normed_params, normed_data

    def makeIDs(self):
        self.normed_name, self.normed_params, self.normed_data = self.makeNormedName()

    def matchesDatatag(self, tagList):
        if self.datatag in tagList or self.normed_data in tagList:
            return True
        return self.datatag.replace('_post', '') in [tag.replace('_post', '') for tag in
                                                     tagList]

    def hasParam(self, name):
        if isinstance(name, str):
            return name in self.param_set
        else:
            return any([True for i in name if i in self.param_set])

    def importanceJobs(self):
        return self.importanceItems

    def importanceJobsRecursive(self):
        res = copy.copy(self.importanceItems)
        for r in self.importanceItems:
            res += r.importanceJobsRecursive()
        return res

    def removeImportance(self, job):
        if job in self.importanceItems:
            self.importanceItems.remove(job)
        else:
            for j in self.importanceItems:
                j.removeImportance(job)

    def makeChainPath(self):
        makePath(self.chainPath)
        return self.chainPath

    def writeIniLines(self, f):
        outfile = open(self.iniFile(), 'w', encoding="utf-8")
        outfile.write("\n".join(f))
        outfile.close()

    def chainName(self, chain=1):
        return self.chainRoot + '_' + str(chain) + '.txt'

    def chainExists(self, chain=1):
        fname = self.chainName(chain)
        return nonEmptyFile(fname)

    def chainNames(self, num_chains=None):
        if num_chains:
            return [self.chainName(i) for i in range(num_chains)]
        else:
            i = 1
            chains = []
            while self.chainExists(i):
                chains.append(self.chainName(i))
                i += 1
            return chains

    def allChainExists(self, num_chains):
        return all([self.chainExists(i + 1) for i in range(num_chains)])

    def chainFileDate(self, chain=1):
        return os.path.getmtime(self.chainName(chain))

    def chainsDodgy(self, interval=600):
        dates = []
        i = 1
        while os.path.exists(self.chainName(i)):
            dates.append(os.path.getmtime(self.chainName(i)))
            i += 1
        return os.path.exists(self.chainName(i + 1)) or max(dates) - min(dates) > interval

    def notRunning(self):
        if not self.chainExists():
            return False  # might be in queue
        lastWrite = self.chainFileDate()
        return lastWrite < time.time() - 5 * 60

    def chainMinimumExists(self):
        fname = self.chainRoot + '.minimum'
        return nonEmptyFile(fname)

    def chainBestfit(self, paramNameFile=None):
        bf_file = self.chainRoot + '.minimum'
        if nonEmptyFile(bf_file):
            return types.BestFit(bf_file, paramNameFile)
        return None

    def chainMinimumConverged(self):
        bf = self.chainBestfit()
        if bf is None:
            return False
        return bf.logLike < 1e29

    def convergeStat(self):
        fname = self.chainRoot + '.converge_stat'
        if not nonEmptyFile(fname):
            return None, None
        textFileHandle = open(fname, encoding="utf-8")
        textFileLines = textFileHandle.readlines()
        textFileHandle.close()
        return (float(textFileLines[0].strip()), len(textFileLines) > 1
                and textFileLines[1].strip() == 'Done')

    def chainFinished(self):
        if self.isImportanceJob:
            done = self.parent.convergeStat()[1]
            if done is None or self.parentChanged() or not self.notRunning():
                return False
        else:
            done = self.convergeStat()[1]
        if done is None:
            return False
        return done

    def wantCheckpointContinue(self, minR=0):
        R, done = self.convergeStat()
        if R is None:
            return False
        if not os.path.exists(self.chainRoot + '_1.chk'):
            return False
        return not done and R > minR

    def getDistExists(self):
        return os.path.exists(self.distRoot + '.margestats')

    def getDistNeedsUpdate(self):
        return self.chainExists() and (not self.getDistExists()
                                       or self.chainFileDate() >
                                       os.path.getmtime(self.distRoot + '.margestats'))

    def parentChanged(self):
        return (not self.chainExists() or
                self.chainFileDate() < self.parent.chainFileDate())

    def R(self):
        if self.result_converge is None:
            fname = self.distRoot + '.converge'
            if not nonEmptyFile(fname):
                return None
            self.result_converge = types.ConvergeStats(fname)
        return float(self.result_converge.worstR())

    def hasConvergeBetterThan(self, R, returnNotExist=False):
        try:
            chainR = self.R()
            if chainR is None:
                return returnNotExist
            return chainR <= R
        except:
            print('WARNING: Bad .converge for ' + self.name)
            return returnNotExist

    def loadJobItemResults(self, paramNameFile=None, bestfit=True, bestfitonly=False,
                           noconverge=False, silent=False):
        self.result_converge = None
        self.result_marge = None
        self.result_likemarge = None
        self.result_bestfit = self.chainBestfit(paramNameFile)
        if not bestfitonly:
            marge_root = self.distRoot
            if self.getDistExists():
                if not noconverge:
                    self.result_converge = types.ConvergeStats(
                        marge_root + '.converge')
                self.result_marge = types.MargeStats(marge_root + '.margestats',
                                                     paramNameFile)
                self.result_likemarge = types.LikeStats(marge_root + '.likestats')
                if self.result_bestfit is not None and bestfit:
                    self.result_marge.addBestFit(self.result_bestfit)
            elif not silent:
                print('missing: ' + marge_root)

    def getMCSamples(self, ini=None, settings=None):
        return loadMCSamples(self.chainRoot, jobItem=self, ini=ini, settings=settings)


class BatchJob(PropertiesItem):
    def __init__(self, path):
        self.batchPath = path
        self.skip = []
        self.subBatches = []
        # self.jobItems = None
        self.getdist_options = {}
        self.iniFile_path = input_folder
        self.scriptFile_path = script_folder
        self.logFile_path = log_folder

    def propertiesIniFile(self):
        return os.path.join(self.batchPath, 'config', 'config.ini')

    def makeItems(self, settings, messages=True):
        self.jobItems = []
        self.getdist_options = getattr(settings, 'getdist_options', self.getdist_options)
        allImportance = getattr(settings, 'importanceRuns', [])
        for group_name, group in settings["grid"]["groups"].items():
            for data_set in group["datasets"]:
                for param_set in group["models"]:
                    if any(data_set in (x.get("skip", {}) or {}).get(param_set, {})
                           for x in (settings["grid"], group)):
                        continue
                    item = JobItem(self.batchPath, param_set, data_set, base=group_name)
                    if hasattr(group, 'groupName'):
                        item.group = group.groupName
                    if item.name not in self.skip:
                        item.makeImportance(group.get("importanceRuns", []))
                        item.makeImportance(allImportance)
                        self.jobItems.append(item)
        for item in getattr(settings, 'jobItems', []):
            self.jobItems.append(item)
            item.makeImportance(allImportance)
        if hasattr(settings, 'importance_filters'):
            for job in self.jobItems:
                for item in job.importanceJobs():
                    item.makeImportance(settings.importance_filters)
                job.makeImportance(settings.importance_filters)
        for item in list(self.items()):
            for x in [imp for imp in item.importanceJobsRecursive()]:
                if self.has_normed_name(x.normed_name):
                    if messages:
                        print('replacing importance sampling run '
                              'with full run: %s' % x.name)
                    item.removeImportance(x)
        for item in list(self.items()):
            for x in [imp for imp in item.importanceJobsRecursive()]:
                if self.has_normed_name(x.normed_name, wantImportance=True, exclude=x):
                    if messages:
                        print('removing duplicate importance sampling run: ' + x.name)
                    item.removeImportance(x)

    def items(self, wantSubItems=True, wantImportance=False):
        for item in self.jobItems:
            yield item
            if wantImportance:
                for imp in item.importanceJobsRecursive():
                    if imp.name not in self.skip:
                        yield imp
        if wantSubItems:
            for subBatch in self.subBatches:
                for item in subBatch.items(wantSubItems, wantImportance):
                    yield item

    def hasName(self, name, wantSubItems=True):
        for jobItem in self.items(wantSubItems):
            if jobItem.name == name:
                return True
        return False

    def has_normed_name(self, name, wantSubItems=True, wantImportance=False,
                        exclude=None):
        return self.normed_name_item(name, wantSubItems, wantImportance,
                                     exclude) is not None

    def normed_name_item(self, name, wantSubItems=True, wantImportance=False,
                         exclude=None):
        for jobItem in self.items(wantSubItems, wantImportance):
            if jobItem.normed_name == name and jobItem is not exclude:
                return jobItem
        return None

    def normalizeDataTag(self, tag):
        return "_".join(sorted(tag.replace('_post', '').split('_')))

    def resolveName(self, paramtag, datatag, wantSubItems=True, wantImportance=True,
                    raiseError=True, base='base',
                    returnJobItem=False):
        if paramtag:
            if isinstance(paramtag, str):
                paramtag = paramtag.split('_')
            paramtags = [base] + sorted(paramtag)
        else:
            paramtag = [base]
            paramtags = [base]
        name = "_".join(paramtags) + '_' + self.normalizeDataTag(datatag)
        jobItem = self.normed_name_item(name, wantSubItems, wantImportance)
        if jobItem is not None:
            return (jobItem.name, jobItem)[returnJobItem]
        if raiseError:
            raise Exception('No match for paramtag, datatag... ' + "_".join(
                paramtag) + ', ' + datatag)
        else:
            return None

    def resolveRoot(self, root):
        for jobItem in self.items(True, True):
            if jobItem.name == root:
                return jobItem
        return self.normed_name_item(root, True, True)

    def save(self, filename=''):
        saveobject(self, (grid_cache_file(self.batchPath), filename)[filename != ''])

    def makeDirectories(self, setting_file=None):
        makePath(self.batchPath)
        if setting_file:
            makePath(self.batchPath + 'config')
            setting_file = setting_file.replace('.pyc', '.py')
            shutil.copy(setting_file, self.batchPath + 'config')
            props = self.propertiesIni()
            props.params['setting_file'] = os.path.split(setting_file)[-1]
            props.saveFile()
        makePath(self.batchPath + self.iniFile_path)
        makePath(self.batchPath + self.scriptFile_path)
        makePath(self.batchPath + self.logFile_path)



"""
Tests some Prior methods.
"""

import numpy as np

from cobaya.parameterization import Parameterization
from cobaya.prior import Prior


def test_prior_confidence():
    info_params = {
        "a": {"prior": {"dist": "uniform", "min": 0, "max": 1}},
        "b": {"prior": {"dist": "norm", "loc": 0, "scale": 1}},
        "c": {"prior": {"dist": "beta", "min": 0, "max": 1, "a": 2, "b": 5}},
    }
    p = Prior(Parameterization(info_params))
    test_confidence_p1 = np.array(
        [[0.45, 0.55], [-0.12566135, 0.12566135], [0.24325963, 0.28641175]])
    assert np.allclose(p.bounds(confidence=0.1), test_confidence_p1)
    test_bounds_p68 = np.array([[0., 1.], [-0.99445788, 0.99445788], [0., 1.]])
    assert np.allclose(p.bounds(confidence_for_unbounded=0.68), test_bounds_p68)



from cobaya.likelihoods.base_classes import SN


class pantheon(SN):
    r"""
    Likelihood of the Pantheon type Ia supernova sample \cite{Scolnic:2017caz},
    including data from the Pan-STARRS1 (PS1) Medium Deep Survey.
    """
    pass



"""
.. module:: planck_2018_CamSpec

:Synopsis: Definition of python-native CamSpec 2018 likelihood (not official Planck product)
:Author: Antony Lewis (from CamSpec f90 source by GPE, StG and AL)

This python version loads the covariance, and cuts it as requested (then inverting). It
can take a few seconds the first time it is loaded, but the inverse can be cached.
The Planck likelihood code (clik) is not required.

Use dataset_params : { 'use_cl': '100x100 143x143 217x217 143x217'} to use e.g. just TT ,
or other combination with TE and EE.

Set use_range to string representation of L range to use, e.g. 50-100, 200-500, 1470-2500,
or pass a dictionary of ranges for each spectrum.

"""

# Global
import numpy as np
import os
import scipy

# Local
from cobaya.likelihoods.base_classes import DataSetLikelihood

use_cache = True


def range_to_ells(use_range):
    """splits range string like '2-5 7 15-3000' into list of specific numbers"""

    if isinstance(use_range, str):
        ranges = []
        for ell_range in use_range.split():
            if '-' in ell_range:
                mn, mx = [int(x) for x in ell_range.split('-')]
                ranges.append(range(mn, mx + 1))
            else:
                ranges.append(int(ell_range))
        return np.concatenate(ranges)
    else:
        return use_range


class Planck2018CamSpecPython(DataSetLikelihood):
    install_options = {"github_repository": "CobayaSampler/planck_native_data",
                       "github_release": "v1", "asset": "CamSpec2018.zip",
                       "directory": "planck_2018_CamSpec_native"}

    type = "CMB"

    @classmethod
    def get_bibtex(cls):
        from cobaya.likelihoods.base_classes import Planck2018Clik
        return Planck2018Clik.get_bibtex()

    def read_normalized(self, filename, pivot=None):
        # arrays all based at L=0, in L(L+1)/2pi units
        self.mpi_debug('Loading: %s', filename)
        dat = np.loadtxt(filename)
        assert int(dat[0, 0]) == 2
        dat = np.hstack(([0, 0], dat[:, 1]))
        if pivot is not None:
            assert pivot < dat.shape[0] + 2
            dat /= dat[pivot]
        return dat

    def init_params(self, ini, silent=False):
        spectra = np.loadtxt(ini.relativeFileName('cl_hat_file'))
        covmat_cl = ini.split('covmat_cl')
        self.use_cl = ini.split('use_cl', covmat_cl)
        if ini.hasKey('use_range'):
            used_ell = ini.params['use_range']
            if isinstance(used_ell, dict):
                self.mpi_info('Using range %s', used_ell)
                used_ell = {key: range_to_ells(value) for key, value in used_ell.items()}
            else:
                if silent:
                    self.mpi_info('CamSpec using range: %s', used_ell)
                used_ell = range_to_ells(used_ell)
        else:
            used_ell = None
        data_vector = []
        nX = 0
        used_indices = []
        with open(ini.relativeFileName('data_ranges'), "r", encoding="utf-8-sig") as f:
            lines = f.readlines()
            while not lines[-1].strip():
                lines = lines[:-1]
            self.Nspec = len(lines)
            lmin = np.zeros(self.Nspec, dtype=int)
            lmax = np.zeros(self.Nspec, dtype=int)
            self.cl_names = []
            self.ell_ranges = np.empty(self.Nspec, dtype=object)
            self.used_sizes = np.zeros(self.Nspec, dtype=int)
            for i, line in enumerate(lines):
                items = line.split()
                tp = items[0]
                self.cl_names.append(tp)
                lmin[i], lmax[i] = [int(x) for x in items[1:]]
                if lmax[i] and lmax[i] >= lmin[i]:
                    n = lmax[i] - lmin[i] + 1
                    data_vector.append(spectra[lmin[i]:lmax[i] + 1, i])
                    if tp in self.use_cl:
                        if used_ell is not None and (
                                not isinstance(used_ell, dict) or tp in used_ell):
                            if isinstance(used_ell, dict):
                                ells = used_ell[tp]
                            else:
                                ells = used_ell
                            self.ell_ranges[i] = np.array(
                                [L for L in range(lmin[i], lmax[i] + 1) if L in ells],
                                dtype=int)
                            used_indices.append(self.ell_ranges[i] + (nX - lmin[i]))
                        else:
                            used_indices.append(range(nX, nX + n))
                            self.ell_ranges[i] = range(lmin[i], lmax[i] + 1)
                        self.used_sizes[i] = len(self.ell_ranges[i])
                    else:
                        lmax[i] = -1
                    nX += n

        self.cl_used = np.array([name in self.use_cl for name in self.cl_names],
                                dtype=bool)
        covfile = ini.relativeFileName('covmat_fiducial')
        with open(covfile, "rb") as cov_f:
            cov = np.fromfile(cov_f, dtype=[np.float32, np.float64]['64.bin' in covfile])
        assert (nX ** 2 == cov.shape[0])
        used_indices = np.concatenate(used_indices)
        self.data_vector = np.concatenate(data_vector)[used_indices]
        self.cov = cov.reshape(nX, nX)[np.ix_(used_indices, used_indices)].astype(
            np.float64)
        if not silent:
            for name, mn, mx in zip(self.cl_names, lmin, lmax):
                if name in self.use_cl:
                    self.mpi_info("L-range for %s: %s %s", name, mn, mx)
            self.mpi_info('Number of data points: %s', self.cov.shape[0])
        self.lmax = lmax
        self.lmin = lmin
        max_l = np.max(self.lmax)
        self.ls = np.arange(max_l + 1)
        self.llp1 = self.ls * (self.ls + 1)

        if np.any(self.cl_used[:4]):
            pivot = 3000
            self.sz_143 = self.read_normalized(
                ini.relativeFileName('sz143file'), pivot)[:max_l + 1]
            self.ksz = self.read_normalized(
                ini.relativeFileName('kszfile'), pivot)[:max_l + 1]
            self.tszxcib = self.read_normalized(
                ini.relativeFileName('tszxcibfile'), pivot)[:max_l + 1]

            self.cib_217 = self.read_normalized(
                ini.relativeFileName('cib217file'), pivot)[:max_l + 1]

            self.dust = np.vstack(
                (self.read_normalized(ini.relativeFileName('dust100file'))[:max_l + 1],
                 self.read_normalized(ini.relativeFileName('dust143file'))[:max_l + 1],
                 self.read_normalized(ini.relativeFileName('dust217file'))[:max_l + 1],
                 self.read_normalized(ini.relativeFileName('dust143x217file'))[
                 :max_l + 1]))
            self.lnrat = self.ls * 0
            l_min = np.min(lmin[self.cl_used])
            self.lnrat[l_min:] = np.log(self.ls[l_min:] / np.float64(pivot))

        import hashlib
        cache_file = self.dataset_filename.replace('.dataset',
                                                   '_covinv_%s.npy' % hashlib.md5(
                                                       str(ini.params).encode(
                                                           'utf8')).hexdigest())
        if use_cache and os.path.exists(cache_file):
            self.covinv = np.load(cache_file).astype(np.float64)
        else:
            self.covinv = np.linalg.inv(self.cov)
            if use_cache:
                np.save(cache_file, self.covinv.astype(np.float32))

    def get_foregrounds(self, data_params):

        sz_bandpass100_nom143 = 2.022
        cib_bandpass143_nom143 = 1.134
        sz_bandpass143_nom143 = 0.95
        cib_bandpass217_nom217 = 1.33

        Aps = np.empty(4)
        Aps[0] = data_params['aps100']
        Aps[1] = data_params['aps143']
        Aps[2] = data_params['aps217']
        Aps[3] = data_params['psr'] * np.sqrt(Aps[1] * Aps[2])
        Aps *= 1e-6 / 9  # scaling convention

        Adust = np.atleast_2d(
            [data_params['dust100'], data_params['dust143'], data_params['dust217'],
             data_params['dust143x217']]).T

        acib143 = data_params.get('acib143', -1)
        acib217 = data_params['acib217']
        cibr = data_params['cibr']
        ncib = data_params['ncib']
        cibrun = data_params['cibrun']

        asz143 = data_params['asz143']
        xi = data_params['xi']
        aksz = data_params['aksz']

        lmax = np.max(self.lmax)

        cl_cib = np.exp(ncib * self.lnrat + cibrun * self.lnrat ** 2 / 2) * self.cib_217
        if acib143 < 0:
            # fix 143 from 217
            acib143 = .094 * acib217 / cib_bandpass143_nom143 * cib_bandpass217_nom217
            # The above came from ratioing Paolo's templates, which were already
            # colour-corrected, and assumed perfect correlation

        ksz = aksz * self.ksz
        C_foregrounds = np.empty((4, lmax + 1))
        # 100
        C_foregrounds[0, :] = ksz + asz143 * sz_bandpass100_nom143 * self.sz_143

        # 143
        A_sz_143_bandpass = asz143 * sz_bandpass143_nom143
        A_cib_143_bandpass = acib143 * cib_bandpass143_nom143
        zCIB = A_cib_143_bandpass * cl_cib
        C_foregrounds[1, :] = (zCIB + ksz + A_sz_143_bandpass * self.sz_143
                               - 2.0 * np.sqrt(
                    A_cib_143_bandpass * A_sz_143_bandpass) * xi * self.tszxcib)

        # 217
        A_cib_217_bandpass = acib217 * cib_bandpass217_nom217
        zCIB = A_cib_217_bandpass * cl_cib
        C_foregrounds[2, :] = zCIB + ksz

        # 143x217
        zCIB = np.sqrt(A_cib_143_bandpass * A_cib_217_bandpass) * cl_cib
        C_foregrounds[3, :] = (cibr * zCIB + ksz - np.sqrt(
            A_cib_217_bandpass * A_sz_143_bandpass) * xi * self.tszxcib)

        # Add dust and point sources
        C_foregrounds += Adust * self.dust + np.outer(Aps, self.llp1)

        return C_foregrounds

    def get_cals(self, data_params):
        calPlanck = data_params.get('A_planck', 1) ** 2
        cal0 = data_params.get('cal0', 1)
        cal2 = data_params.get('cal2', 1)
        calTE = data_params.get('calTE', 1)
        calEE = data_params.get('calEE', 1)
        return np.array([cal0, 1, cal2, np.sqrt(cal2), calTE, calEE]) * calPlanck

    def chi_squared(self, CT, CTE, CEE, data_params):

        cals = self.get_cals(data_params)
        if np.any(self.cl_used[:4]):
            foregrounds = self.get_foregrounds(data_params)
        delta_vector = self.data_vector.copy()
        ix = 0
        for i, (cal, n) in enumerate(zip(cals, self.used_sizes)):
            if n > 0:
                if i <= 3:
                    # noinspection PyUnboundLocalVariable
                    delta_vector[ix:ix + n] -= (CT[self.ell_ranges[i]] +
                                                foregrounds[i][self.ell_ranges[i]]) / cal
                elif i == 4:
                    delta_vector[ix:ix + n] -= CTE[self.ell_ranges[i]] / cal
                elif i == 5:
                    delta_vector[ix:ix + n] -= CEE[self.ell_ranges[i]] / cal
                ix += n
        return self._fast_chi_squared(self.covinv, delta_vector)

    def logp(self, **data_params):
        Cls = self.provider.get_Cl(ell_factor=True)
        return -0.5 * self.chi_squared(Cls.get('tt'), Cls.get('te'), Cls.get('ee'),
                                       data_params)

    def get_requirements(self):
        # State requisites to the theory code
        l_max = np.max(self.lmax)
        used = []
        if np.any(self.cl_used[:4]):
            used += ['tt']
        if 'TE' in self.use_cl:
            used += ['te']
        if 'EE' in self.use_cl:
            used += ['ee']
        return {"Cl": {cl: l_max for cl in used}}

    def coadded_TT(self, data_params=None, foregrounds=None, cals=None,
                   want_cov=True, data_vector=None):
        nTT = np.sum(self.used_sizes[:4])
        assert nTT
        if foregrounds is not None and cals is not None and data_params is not None:
            raise ValueError('data_params not used')
        if foregrounds is None:
            assert data_params is not None
            foregrounds = self.get_foregrounds(data_params)
        if cals is None:
            assert data_params is not None
            cals = self.get_cals(data_params)
        if data_vector is None:
            data_vector = self.data_vector
        delta_vector = data_vector[:nTT].copy()
        cal_vector = np.zeros(delta_vector.shape)
        lmin = np.min([min(r) for r in self.ell_ranges[:4]])
        lmax = np.max([max(r) for r in self.ell_ranges[:4]])
        n_p = lmax - lmin + 1
        LS = np.zeros(delta_vector.shape, dtype=int)
        ix = 0
        for i, (cal, n) in enumerate(zip(cals[:4], self.used_sizes[:4])):
            if n > 0:
                delta_vector[ix:ix + n] -= foregrounds[i][self.ell_ranges[i]] / cal
                LS[ix:ix + n] = self.ell_ranges[i]
                cal_vector[ix:ix + n] = cal
                ix += n
        pcov = np.zeros((n_p, n_p))
        d = self.covinv[:nTT, :nTT].dot(delta_vector)
        dL = np.zeros(n_p)
        ix1 = 0
        ell_offsets = [LS - lmin for LS in self.ell_ranges[:4]]
        contiguous = not np.any(np.count_nonzero(LS - np.arange(LS[0],
                                                                LS[-1] + 1, dtype=int))
                                for LS in self.ell_ranges[:4])
        for i, (cal, LS, n) in enumerate(zip(cals[:4], ell_offsets, self.used_sizes[:4])):
            dL[LS] += d[ix1:ix1 + n] / cal
            ix = 0
            for cal2, r in zip(cals[:4], ell_offsets):
                if contiguous:
                    pcov[LS[0]:LS[0] + n, r[0]:r[0] + len(r)] += \
                        self.covinv[ix1:ix1 + n, ix:ix + len(r)] / (cal2 * cal)
                else:
                    pcov[np.ix_(LS, r)] += \
                        self.covinv[ix1:ix1 + n, ix:ix + len(r)] / (cal2 * cal)
                ix += len(r)
            ix1 += n

        CTot = np.zeros(self.ls[-1] + 1)
        if want_cov:
            pcovinv = np.linalg.inv(pcov)
            CTot[lmin:lmax + 1] = pcovinv.dot(dL)
            return CTot, pcovinv
        else:
            try:
                CTot[lmin:lmax + 1] = scipy.linalg.solve(pcov, dL, assume_a='pos')
            except:
                CTot[lmin:lmax + 1] = np.linalg.solve(pcov, dL)
            return CTot

    def get_weights(self, data_params):
        # get weights for each temperature spectrum as function of L
        ix = 0
        f = self.get_foregrounds(data_params) * 0
        weights = []
        for i in range(4):
            ells = self.ell_ranges[i]
            vec = np.zeros(self.data_vector.shape)
            vec[ix:ix + len(ells)] = 1
            Ti = self.coadded_TT(data_params, data_vector=vec, want_cov=False,
                                 foregrounds=f)
            weights.append((ells, Ti[ells]))
            ix += len(ells)
        return weights

    def diff(self, spec1, spec2, data_params):
        """
        Get difference (residual) between frequency spectra and the covariance
        :param spec1: name of spectrum 1
        :param spec2:  name of spectrum 2
        :param data_params: dictionary of parameters
        :return: ell range array, difference array, covariance matrix
        """
        foregrounds = self.get_foregrounds(data_params)
        cals = self.get_cals(data_params)
        i = self.cl_names.index(spec1)
        j = self.cl_names.index(spec2)
        off1 = np.sum(self.used_sizes[:i])
        off2 = np.sum(self.used_sizes[:j])
        lmax = np.min([max(r) for r in self.ell_ranges[[i, j]]])
        lmin = np.max([min(r) for r in self.ell_ranges[[i, j]]])

        diff = np.zeros(self.ls[-1] + 1)
        diff[self.ell_ranges[i]] = (self.data_vector[off1:off1 + self.used_sizes[i]] *
                                    cals[i] - foregrounds[i][self.ell_ranges[i]])
        diff[self.ell_ranges[j]] -= (
                self.data_vector[off2:off2 + self.used_sizes[j]] * cals[j] -
                foregrounds[j][
                    self.ell_ranges[j]])
        cov = self.cov
        n_p = lmax - lmin + 1
        off1 += lmin - np.min(self.ell_ranges[i])
        off2 += lmin - np.min(self.ell_ranges[j])
        pcov = cals[i] ** 2 * cov[off1:off1 + n_p, off1:off1 + n_p] \
               + cals[j] ** 2 * cov[off2:off2 + n_p, off2:off2 + n_p] \
               - cals[i] * cals[j] * (
                       cov[off2:off2 + n_p, off1:off1 + n_p] + cov[off1:off1 + n_p,
                                                               off2:off2 + n_p])
        return range(lmin, lmax + 1), diff[lmin:lmax + 1], pcov



# Minimization of a random Gaussian likelihood using the minimize sampler.

import numpy as np
import pytest
import os

from cobaya import mpi, run, InputDict, Likelihood
from cobaya.samplers.minimize import valid_methods

pytestmark = pytest.mark.mpi

mean = np.array([0.30245268, 0.61884443, 0.5])
cov = np.array(
    [[0.00796336, -0.0014805, -0.00479433], [-0.0014805, 0.00561415, 0.00434189],
     [-0.00479433, 0.00434189, 0.03208593]])
_inv_cov = np.linalg.inv(cov)


class NoisyCovLike(Likelihood):
    params = {'a': [0, 1, 0.5, 0.3, 0.08], 'b': [0, 1, 0.5, 0.3, 0.08],
              'c': [0, 1, 0.5, 0.3, 0.08]}
    noise = 0

    def logp(self, **params_values):
        x = np.array([params_values['a'], params_values['b'], params_values['c']]) - mean
        return -_inv_cov.dot(x).dot(x) / 2 + self.noise * np.random.random_sample()


@mpi.sync_errors
def test_minimize_gaussian(tmpdir):
    maxloglik = 0
    for method in reversed(valid_methods):
        NoisyCovLike.noise = 0.005 if method == 'bobyqa' else 0
        info: InputDict = {'likelihood': {'like': NoisyCovLike},
                           "sampler": {"minimize": {"ignore_prior": True,
                                                    "method": method}}}
        products = run(info)[1].products()
        error = abs(maxloglik - -products["minimum"]["minuslogpost"])
        assert error < 0.01

        info['output'] = os.path.join(tmpdir, 'testmin')
        products = run(info, force=True)[1].products()
        if mpi.is_main_process():
            from getdist.types import BestFit
            res = BestFit(info['output'] + '.bestfit').getParamDict()
            assert np.isclose(res["loglike"], products["minimum"]["minuslogpost"])
            for p, v in list(res.items())[:-2]:
                assert np.isclose(products["minimum"][p], v)


@mpi.sync_errors
def test_run_minimize(tmpdir):
    NoisyCovLike.noise = 0
    info: InputDict = {'likelihood': {'like': NoisyCovLike},
                       "sampler": {"mcmc": {"Rminus1_stop": 0.5, 'Rminus1_cl_stop': 0.4,
                                            'seed': 2}},
                       "output": os.path.join(tmpdir, 'testchain')}
    run(info, force=True)
    min_info: InputDict = dict(info, sampler={'minimize': None})
    output_info, sampler = run(min_info, force=True)
    assert (abs(sampler.products()["minimum"]["b"] - mean[1]) < 0.01)



from cobaya.likelihoods.base_classes import Planck2018Clik


class TT_unbinned(Planck2018Clik):
    r"""
    Unbinned high-$\ell$ temperature-only \textsc{plik} likelihood of Planck's 2018 data
    release \cite{Aghanim:2019ame}.
    """
    pass



import os
import numpy as np
from cobaya.likelihoods.base_classes import InstallableLikelihood


class EE(InstallableLikelihood):
    """
    Python translation of the Planck 2018 SimALl EE likelihood (python AL, Oct 2022)
    See https://wiki.cosmos.esa.int/planck-legacy-archive/index.php/CMB_spectrum_%26_Likelihood_Code
    Equivalent to planck_2018_lowl.EE: the public Planck clik likelihood using file
    simall_100x143_offlike5_EE_Aplanck_B.clik (data converted to prob_table.txt)

    Calibration error is very small compared to EE uncertainty, but calibration can be
    used; otherwise taken to be 1
    """
    install_options = {"github_repository": "CobayaSampler/planck_native_data",
                       "github_release": "v1", "asset": "planck_2018_lowE.zip",
                       "directory": "planck_2018_lowE_native"}
    type = "CMB"
    aliases = ["lowE"]

    _lmin = 2
    _lmax = 29
    _nstepsEE = 3000
    _stepEE = 0.0001

    @classmethod
    def get_bibtex(cls):
        from cobaya.likelihoods.base_classes import Planck2018Clik
        return Planck2018Clik.get_bibtex()

    def initialize(self):
        if self.get_install_options() and self.packages_path:
            path = self.get_path(self.packages_path)
            self.probEE = np.loadtxt(os.path.join(path, 'prob_table.txt'))

    def get_can_support_params(self):
        return ['A_planck']

    def get_requirements(self):
        return {'Cl': {'ee': self._lmax}}

    def log_likelihood(self, cls_EE, calib=1):
        r"""
        Calculate log likelihood from CMB EE spectrum by using likelihood table

        :param cls_EE: L(L+1)C_L/2pi zero-based array in muK^2 units
        :param calib: optional calibration parameter
        :return: log likelihood
        """
        EE_index = (cls_EE[self._lmin:self._lmax + 1]
                    / (calib ** 2 * self._stepEE)).astype(int)
        try:
            return np.take_along_axis(self.probEE, EE_index[np.newaxis, :], 0).sum()
        except IndexError:
            self.log.warning("low EE multipole out of range, rejecting point")
            return -np.inf

    def logp(self, **params_values):
        cls = self.provider.get_Cl(ell_factor=True)['ee']
        return self.log_likelihood(cls, params_values.get('A_planck', 1))



"""
Allows calling `cobaya-[command]` as `python -m cobaya [command]`.

`run` is optional: one can pass directly an input file as
`python -m cobaya input.yaml`.
"""

import sys
from importlib import import_module

commands = {"install": ["install", "install_script"],
            "doc": ["doc", "doc_script"],
            "bib": ["bib", "bib_script"],
            "run": ["run", "run_script"],
            "cosmo-generator": ["cosmo_input", "gui_script"],
            "create-image": ["containers", "create_image_script"],
            "prepare-data": ["containers", "prepare_data_script"],
            "grid-create": ["grid_tools", "make_grid_script"],
            "grid-run": ["grid_tools.runbatch", "run"],
            "run-job": ["grid_tools.runMPI", "run_single"],
            }

help_msg = ("Add a one of the following commands and its arguments "
            "(`<command> -h` for help): %r" % list(commands))

if __name__ == "__main__":

    try:
        command_or_input = sys.argv[1].lower()
    except IndexError:  # no command
        print(help_msg)
        exit()
    else:
        module, func = commands.get(command_or_input, (None, None))

        if module is not None:
            sys.argv.pop(1)
            assert func is not None
            getattr(import_module("cobaya." + module), func)()
        else:
            if command_or_input in ["-h", "--help"]:
                print(help_msg)
                exit()
            else:
                # no command --> assume run with input file as 1st arg (don't pop!)
                module, func = commands["run"]
                getattr(import_module("cobaya." + module), func)(
                    help_commands=str(list(commands)))



"""
.. module:: parameterization

:Synopsis: Class managing the possibly different parameterizations
           used by sampler and likelihoods
:Author: Jesus Torrado

"""
# Global
import numpy as np
from numbers import Real
from itertools import chain
from copy import deepcopy
from typing import Mapping, Sequence, Dict, Set, List, Tuple, Any, Callable, Union

# Local
from cobaya.typing import ParamsDict, ParamDict, ParamInput, \
    ExpandedParamsDict, ParamValuesDict, partags
from cobaya.tools import get_external_function, ensure_nolatex, is_valid_variable_name, \
    getfullargspec, deepcopy_where_possible, invert_dict, str_to_list
from cobaya.log import LoggedError, HasLogger


def is_fixed_or_function_param(info_param: ParamInput) -> bool:
    """
    Returns True if the parameter has been fixed to a value or through a function.
    """
    return expand_info_param(info_param).get("value") is not None


def is_sampled_param(info_param: ParamInput) -> bool:
    """
    Returns True if the parameter has a prior.
    """
    return "prior" in expand_info_param(info_param)


def is_derived_param(info_param: ParamInput) -> bool:
    """
    Returns True if the parameter is saved as a derived one.
    """
    return expand_info_param(info_param).get("derived", False) is not False


def expand_info_param(info_param: ParamInput, default_derived=True) -> ParamDict:
    """
    Expands the info of a parameter, from the user-friendly, shorter format
    to a more unambiguous one.
    """
    info_param = deepcopy_where_possible(info_param)
    if not isinstance(info_param, Mapping):
        if info_param is None:
            info_param = {}
        elif isinstance(info_param, Sequence) and not isinstance(info_param, str):
            values = list(info_param)
            allowed_lengths = [2, 4, 5]
            if len(values) not in allowed_lengths:
                raise LoggedError(__name__,
                                  "Parameter info length not valid: %d. "
                                  "The allowed lengths are %r. See documentation.",
                                  len(values), allowed_lengths)
            info_param = {"prior": [values[0], values[1]]}
            if len(values) >= 4:
                info_param["ref"] = [values[2], values[3]]
                if len(values) == 5:
                    info_param["proposal"] = values[4]
        else:
            info_param = {"value": info_param}
    if all(f not in info_param for f in ["prior", "value", "derived"]):
        info_param["derived"] = default_derived
    # Dynamical input parameters: save as derived by default
    value = info_param.get("value")
    if isinstance(value, str) or callable(value):
        info_param["derived"] = info_param.get("derived", True)
    # noinspection PyTypeChecker
    return info_param


def reduce_info_param(info_param: ParamDict) -> ParamInput:
    """
    Compresses the info of a parameter, suppressing default values.
    This is the opposite of :func:`~input.expand_info_param`.
    """
    info_param = deepcopy_where_possible(info_param)
    if not isinstance(info_param, dict):
        return None
    # All parameters without a prior are derived parameters unless otherwise specified
    if info_param.get("derived") is True:
        info_param.pop("derived")
    # Fixed parameters with single "value" key
    if list(info_param) == ["value"] and not callable(info_param["value"]):
        return info_param["value"]
    return info_param


_WrappedFunc = Tuple[Callable, Dict[str, Any], List[str]]


class Parameterization(HasLogger):
    """
    Class managing parameterization.
    Translates parameter between sampler+prior and likelihood
    """

    def __init__(self, info_params: Union[ParamsDict, ExpandedParamsDict],
                 allow_renames=True,
                 ignore_unused_sampled=False):
        self.set_logger()
        self.allow_renames = allow_renames
        # First, we load the parameters,
        # not caring about whether they are understood by any likelihood.
        # `input` contains the parameters (expected to be) understood by the likelihood,
        #   with its fixed value, its fixing function, or None if their value is given
        #   directly by the sampler.
        self._infos = {}
        self._input: ParamValuesDict = {}
        self._input_funcs = {}
        self._input_args = {}
        self._input_dependencies: Dict[str, Set[str]] = {}
        self._dropped: Set[str] = set()
        self._output: ParamValuesDict = {}
        self._constant: ParamValuesDict = {}
        self._sampled: ParamValuesDict = {}
        self._sampled_renames: Dict[str, List[str]] = {}
        self._derived: ParamValuesDict = {}
        self._derived_inputs = []
        self._derived_funcs = {}
        self._derived_args = {}
        self._derived_dependencies: Dict[str, Set[str]] = {}
        # Notice here that expand_info_param *always* adds a "derived":True tag
        # to infos without "prior" or "value", and a "value" field
        # to fixed params
        for p, info in info_params.items():
            if isinstance(info, Mapping) and not set(info).issubset(partags):
                raise LoggedError(self.log, "Parameter '%s' has unknown options %s",
                                  p, set(info).difference(partags))
            info = expand_info_param(info)
            self._infos[p] = info
            if is_fixed_or_function_param(info):
                if isinstance(info["value"], Real):
                    self._constant[p] = float(info["value"])
                    self._input[p] = self._constant[p]
                    if info.get("drop"):
                        self._dropped.add(p)
                else:
                    self._input[p] = np.nan
                    self._input_funcs[p] = get_external_function(info["value"])
                    self._input_args[p] = getfullargspec(self._input_funcs[p]).args
            if is_sampled_param(info):
                self._sampled[p] = np.nan
                self._input[p] = np.nan
                if info.get("drop"):
                    self._dropped.add(p)
                self._sampled_renames[p] = str_to_list(info.get("renames") or [])
            if is_derived_param(info):
                self._derived[p] = np.nan
                # Dynamical parameters whose value we want to save
                if info["derived"] is True and is_fixed_or_function_param(info):
                    # parameters that are already known or computed by input funcs
                    self._derived_inputs.append(p)
                elif info["derived"] is True:
                    self._output[p] = np.nan
                else:
                    self._derived_funcs[p] = get_external_function(info["derived"])
                    self._derived_args[p] = getfullargspec(self._derived_funcs[p]).args
        # Check that the sampled and derived params are all valid python variable names
        for p in chain(self._sampled, self._derived):
            if not is_valid_variable_name(p):
                is_in = p in self._sampled
                eg_in = "  p_prime:\n    prior: ...\n  %s: " \
                        "'lambda p_prime: p_prime'\n" % p
                eg_out = "  p_prime: 'lambda %s: %s'\n" % (p, p)
                raise LoggedError(
                    self.log, "Parameter name '%s' is not a valid Python variable name "
                              "(it needs to start with a letter or '_').\n"
                              "If this is an %s parameter of a likelihood or theory, "
                              "whose name you cannot change,%s define an associated "
                              "%s one with a valid name 'p_prime' as: \n\n%s",
                    p, "input" if is_in else "output",
                    "" if is_in else " remove it and",
                    "sampled" if is_in else "derived",
                    eg_in if is_in else eg_out)

        # input params depend on input and sampled only,
        # never on output/derived unless constant
        known_input = set(self._input)
        all_input_arguments = set(chain(*self._input_args.values()))
        bad_input_dependencies = all_input_arguments - known_input
        if bad_input_dependencies:
            raise LoggedError(
                self.log,
                "Input parameters defined as functions can only depend on other "
                "input parameters. In particular, an input parameter cannot depend on %r."
                " Use an explicit Theory calculator for more complex dependencies.\n"
                "If you intended to define a derived output parameter use derived: "
                "instead of value:", list(bad_input_dependencies))

        # Assume that the *un*known function arguments are likelihood/theory
        # output parameters
        for arg in (all_input_arguments.union(*self._derived_args.values())
                .difference(known_input).difference(self._derived)):
            self._output[arg] = np.nan

        # Useful set: directly "output-ed" derived
        self._directly_output = [p for p in self._derived if p in self._output]

        self._wrapped_input_funcs, self._wrapped_derived_funcs = \
            self._get_wrapped_functions_evaluation_order()

        # Useful mapping: input params that vary if each sample is varied
        self._sampled_input_dependence = {s: [i for i in self._input
                                              if s in self._input_dependencies.get(i, {})]
                                          for s in self._sampled}
        # From here on, some error control.
        # Only actually raise error after checking if used by prior.
        if not ignore_unused_sampled:
            self._dropped_not_directly_used = self._dropped.intersection(
                p for p, v in self._sampled_input_dependence.items() if not v)
        else:
            self._dropped_not_directly_used = set()

        # warn if repeated labels
        labels_inv_repeated = invert_dict(self.labels())
        labels_inv_repeated = {k: v for k, v in labels_inv_repeated.items() if len(v) > 1}
        if labels_inv_repeated:
            self.mpi_warning(
                "There are repeated parameter labels: %r", labels_inv_repeated)

    def dropped_param_set(self) -> Set[str]:
        return self._dropped.copy()

    def input_params(self) -> ParamValuesDict:
        return self._input.copy()

    def output_params(self) -> ParamValuesDict:
        return self._output.copy()

    def constant_params(self) -> ParamValuesDict:
        return self._constant.copy()

    def sampled_params(self) -> ParamValuesDict:
        return self._sampled.copy()

    def sampled_params_info(self) -> ExpandedParamsDict:
        return {p: deepcopy_where_possible(info) for p, info
                in self._infos.items() if p in self._sampled}

    def sampled_params_renames(self) -> Dict[str, List[str]]:
        return deepcopy(self._sampled_renames)

    def derived_params(self) -> ParamValuesDict:
        return self._derived.copy()

    def derived_params_info(self) -> ExpandedParamsDict:
        return {p: deepcopy_where_possible(info) for p, info
                in self._infos.items() if p in self._derived}

    def sampled_input_dependence(self) -> Dict[str, List[str]]:
        return deepcopy(self._sampled_input_dependence)

    def get_input_func(self, p, **params_values):
        func = self._input_funcs[p]
        args = self._input_args[p]
        return func(*[params_values.get(arg) for arg in args])

    @property
    def input_dependencies(self) -> Dict[str, Set[str]]:
        return self._input_dependencies

    def to_input(self, sampled_params_values) -> ParamValuesDict:
        # Gets all current sampled and input derived parameters as a dictionary,
        # including dropped parameters. Result is not a copy and must not be modified.

        # Store sampled params, so that derived can depend on them
        if not isinstance(sampled_params_values, dict):
            sampled_params_values = dict(zip(self._sampled, sampled_params_values))
        else:
            sampled_params_values = sampled_params_values.copy()
        self._sampled = sampled_params_values

        # First include all sampled input parameters,
        self._input.update(sampled_params_values)

        if self._wrapped_input_funcs:
            # Then evaluate the functions
            for p, (func, args, to_set) in self._wrapped_input_funcs.items():
                for arg in to_set:
                    args[arg] = self._input.get(arg, sampled_params_values.get(arg))
                self._input[p] = self._call_param_func(p, func, args)
        return self._input

    def to_derived(self, output_params_values) -> ParamValuesDict:
        if not isinstance(output_params_values, dict):
            output_params_values = dict(zip(self._output, output_params_values))
        # Fill first derived parameters which are direct output parameters
        for p in self._directly_output:
            self._derived[p] = output_params_values[p]
        for p in self._derived_inputs:
            self._derived[p] = self._input[p]
        # Then evaluate the functions
        if self._wrapped_derived_funcs:
            for p, (func, args, to_set) in self._wrapped_derived_funcs.items():
                for arg in to_set:
                    if (val := self._input.get(arg)) is None:
                        if (val := output_params_values.get(arg)) is None:
                            val = self._derived.get(arg)
                    args[arg] = val
                self._derived[p] = self._call_param_func(p, func, args)
        return self._derived

    def check_sampled(self, sampled_params: Union[Sequence[float], Dict[str, float]]
                      ) -> Union[Sequence[float], Dict[str, float]]:
        """
        Performs some checks on the given sampled params.

        If an array is passed, the only test performed is for the right amount of
        parameters, and the same array is returned if successful.

        If a dictionary is passed, it checks that it contains all the sampled parameters,
        and just them. This function is aware of known renamings. Returns dict of
        parameters (model's naming, not remanes) and their values.
        """
        if sampled_params is None:  # only works if there are no sampled params
            sampled_params = []
        if hasattr(sampled_params, "keys"):
            return self.check_sampled_dict(**sampled_params)
        else:
            if len(sampled_params) != len(self._sampled):
                raise LoggedError(self.log, "Wrong number of sampled parameters passed: "
                                            "%d given vs %d expected",
                                  len(sampled_params),
                                  len(self._sampled))
            return sampled_params

    def check_sampled_dict(self, **sampled_params) -> ParamValuesDict:
        """
        Check that the input dictionary contains all the sampled parameters,
        and just them.

        This function is aware of known renamings.

        Returns dict of parameters (model's naming, not renames) and their values.
        """
        sampled_output: ParamValuesDict = {}
        for p, renames in self._sampled_renames.items():
            for pprime in sampled_params:
                if pprime == p or (pprime in renames if self.allow_renames else False):
                    sampled_output[p] = sampled_params.pop(pprime)
                    break
        if len(sampled_output) < len(self._sampled):
            not_found = set(self._sampled).difference(sampled_output)
            if self.allow_renames:
                msg = ("The following expected sampled parameters " +
                       ("(or their aliases) " if self.allow_renames else "") +
                       "were not found : %r",
                       ({p: self._sampled_renames[p] for p in not_found}
                        if self.allow_renames else not_found))
            else:
                msg = ("The following expected sampled parameters "
                       "were not found : %r",
                       {p: self._sampled_renames[p] for p in not_found})
            raise LoggedError(self.log, *msg)
        # Ignore fixed input parameters if they have the correct value
        not_used = set(sampled_params)
        for p, value in sampled_params.items():
            known_value = self._constant.get(p)
            if known_value is None:
                raise LoggedError(self.log, "Unknown parameter %r.", p)
            elif np.allclose(value, known_value):
                not_used.remove(p)
                self.log.debug("Fixed parameter %r ignored.", p)
            else:
                raise LoggedError(
                    self.log, "Cannot change value of constant parameter: "
                              "%s = %g (new) vs %g (old).", p, value, known_value)
        if not_used:
            duplicated = not_used.intersection(
                chain(
                    *[list(chain(*[[k], v])) for k, v in self._sampled_renames.items()]))
            not_used = not_used.difference(duplicated)
            derived = not_used.intersection(self._derived)
            input_ = not_used.intersection(self._input)
            unknown = not_used.difference(derived).difference(input_)
            msg_text = ("Incorrect parameters! " +
                        ("\n   Duplicated entries (using their aliases): %r" %
                         list(duplicated) if duplicated else "") +
                        ("\n   Not known: %r" % list(unknown) if unknown else "") +
                        ("\n   Cannot be fixed: %r " % list(input_) +
                         "--> instead, fix sampled parameters that depend on them!"
                         if input_ else "") +
                        ("\n   Cannot be fixed because are derived parameters: %r " %
                         list(derived) if derived else ""))
            for line in msg_text.split("\n"):
                self.log.error(line)
            raise LoggedError
        return sampled_output

    def check_dropped(self, external_dependence):
        # some error control, given external_dependence from prior
        # only raise error after checking not used by prior
        if self._dropped_not_directly_used.difference(external_dependence):
            raise LoggedError(
                self.log,
                "Parameters %r are sampled but not passed to a likelihood or theory "
                "code, and never used as arguments for any prior or parameter "
                "functions. Check that you are not using "
                "the '%s' tag unintentionally.",
                list(self._dropped_not_directly_used), "drop")

    def labels(self) -> Dict[str, str]:
        """
        Returns a dictionary of LaTeX labels of the sampled and derived parameters.

        Uses the parameter name if no label has been given.
        """

        def get_label(p, info):
            return ensure_nolatex(getattr(info, "get", lambda x, y: y)
                                  ("latex", p.replace("_", r"\ ")))

        return {p: get_label(p, info) for p, info in self._infos.items()}

    def _call_param_func(self, p, func, kwargs):
        try:
            return func(**kwargs)
        except NameError as exception:
            unknown = str(exception).split("'")[1]
            raise LoggedError(
                self.log, "Unknown variable '%s' was referenced in the definition of "
                          "the parameter '%s', with arguments %r.", unknown, p,
                list(kwargs))
        except:
            self.log.error("Function for parameter '%s' failed at evaluation "
                           "and threw the following exception:", p)
            raise

    def _get_wrapped_functions_evaluation_order(self):
        # get evaluation order for input and derived parameter function
        # and pre-prepare argument dicts

        wrapped_funcs: Tuple[Dict[str, _WrappedFunc], Dict[str, _WrappedFunc]] = ({}, {})
        known = set(chain(self._constant, self._sampled))

        for derived, wrapped_func in zip((False, True), wrapped_funcs):
            if derived:
                inputs = self._derived_funcs.copy()
                input_args = self._derived_args
                known.update(self._output)
                output = self._derived
                dependencies = self._derived_dependencies
            else:
                inputs = self._input_funcs.copy()
                input_args = self._input_args
                output = self._input
                dependencies = self._input_dependencies

            while inputs:
                for p, func in inputs.items():
                    args = input_args[p]
                    if not known.issuperset(args):
                        continue
                    known.add(p)
                    dependencies[p] = set(
                        chain(args, *(dependencies.get(arg, []) for arg in args)))

                    if set(args).issubset(self._constant):
                        # all inputs are constant, so output is constant and precomputed
                        self._constant[p] = \
                            self._call_param_func(p, func,
                                                  {arg: self._constant[arg] for arg in
                                                   args})
                        output[p] = self._constant[p]
                    else:
                        # Store function, argument dict with constants pre-filled,
                        # and unset args as tuple
                        wrapped_func[p] = \
                            (func, {arg: self._constant.get(arg) for arg in args},
                             [arg for arg in args if arg not in self._constant])
                    del inputs[p]
                    break
                else:
                    raise LoggedError(
                        self.log, "Could not resolve arguments for parameters %s. "
                                  "Maybe there is a circular dependency between derived "
                                  "parameters?", list(inputs))
        return wrapped_funcs



from typing import Dict, Any, Optional, Union, Sequence, Type, Callable, Mapping
from typing import TypedDict, Literal
from types import MappingProxyType

InfoDict = Dict[str, Any]
InfoDictIn = Mapping[str, Any]

# an immutable empty dict (e.g. for argument defaults)
empty_dict: InfoDictIn = MappingProxyType({})
# empty iterator-compatible object that can be used like None to test if assigned
unset_params: Sequence[str] = ()

LikeDict = InfoDict
TheoryDict = InfoDict
SamplerDict = InfoDict

# Read-only versions for input
LikeDictIn = InfoDictIn
TheoryDictIn = InfoDictIn
SamplerDictIn = InfoDictIn

ParamValuesDict = Dict[str, float]
# Do not yet explicitly support passing instances here
TheoriesDict = Dict[str, Union[None, TheoryDict, Type]]
LikesDict = Dict[str, Union[None, LikeDict, Type, Callable]]
SamplersDict = Dict[str, Optional[SamplerDict]]
PriorsDict = Dict[str, Union[str, Callable]]

# parameters in a params list can be specified on input by
# 1. a ParamDict dictionary
# 2. constant value
# 3. a string giving lambda function of other parameters
# 4. None - must be a computed output parameter
# 5. Sequence specifying uniform prior range [min, max] and optionally
#    'ref' mean and standard deviation for starting positions, and optionally
#    proposal width. Allowed lengths, 2, 4, 5
ParamInput = Union['ParamDict', None, str, float, Sequence[float]]
ParamsDict = Dict[str, ParamInput]
ExpandedParamsDict = Dict[str, 'ParamDict']

partags = {"prior", "ref", "proposal", "value", "drop",
           "derived", "latex", "renames", "min", "max"}

LiteralFalse = Literal[False]
ModelBlock = Literal["theory", "likelihood", "prior", "params"]
Kind = Literal["theory", "likelihood", "sampler"]


class SciPyDistDict(TypedDict):
    dist: str
    loc: float
    scale: float


class SciPyMinMaxDict(TypedDict, total=False):
    dist: str  # default uniform
    min: float
    max: float


class ParamDict(TypedDict, total=False):
    value: Union[float, Callable, str]
    derived: Union[bool, str, Callable]
    prior: Union[None, Sequence[float], SciPyDistDict, SciPyMinMaxDict]
    ref: Union[None, Sequence[float], SciPyDistDict, SciPyMinMaxDict]
    proposal: Optional[float]
    renames: Union[str, Sequence[str]]
    latex: str
    drop: bool  # true if parameter should not be available for assignment to theories
    min: float  # hard bounds (does not affect prior)
    max: float


class ModelDict(TypedDict, total=False):
    theory: TheoriesDict
    likelihood: LikesDict
    prior: PriorsDict
    params: ParamsDict
    auto_params: ParamsDict


class PostDict(TypedDict, total=False):
    add: Optional[ModelDict]
    remove: Union[None, ModelDict, Dict[str, Union[str, Sequence[str]]]]
    output: Optional[str]
    suffix: Optional[str]
    skip: Union[None, float, int]
    thin: Optional[int]
    packages_path: Optional[str]


class InputDict(ModelDict, total=False):
    sampler: SamplersDict
    post: PostDict
    force: bool
    debug: Union[bool, int, str]
    resume: bool
    stop_at_error: bool
    test: bool
    timing: bool
    packages_path: Optional[str]
    output: Optional[str]
    version: Optional[Union[str, InfoDict]]



"""
Tests the setting and updating of the reference pdf, including re-checking if point-like.
"""
import pytest
import numpy as np

from cobaya.model import get_model
from cobaya.sampler import get_sampler
from cobaya import mpi


def test_ref():
    val = 1
    mean, std = 0.5, 0.1
    info = {
        "params": {
            "a": {"prior": [0, 1]},
            "b": {"prior": [0, 1], "ref": None},
            "c": {"prior": [0, 1], "ref": val},
            "d": {"prior": [0, 1], "ref": [mean, std]},
            "e": {"prior": [0, 1], "ref": {"dist": "norm", "loc": mean, "scale": std}}},
        "likelihood": {"one": None}}
    model = get_model(info)
    for i in [3, 4]:
        assert model.prior.ref_pdf[i].dist.name == "norm"
        assert model.prior.ref_pdf[i].mean() == mean
        assert model.prior.ref_pdf[i].std() == std
    # Not point-like: 2 nan's (use prior) and 2 norms
    assert not model.prior.reference_is_pointlike
    # Let's update it -- remove norms
    upd1_ref_info = {
        "d": val + 2,
        "e": val + 3}
    model.prior.set_reference(upd1_ref_info)
    # Updated:
    assert model.prior.ref_pdf[3] == upd1_ref_info["d"]
    assert model.prior.ref_pdf[4] == upd1_ref_info["e"]
    # Unchanged:
    assert model.prior.ref_pdf[0] is np.nan
    assert model.prior.ref_pdf[1] is np.nan
    assert model.prior.ref_pdf[2] == val
    # Still not pointlike: there are nan's --> use prior
    assert not model.prior.reference_is_pointlike
    # Let's update it -- remove nan's
    upd2_ref_info = {
        "a": val - 2,
        "b": val - 1}
    model.prior.set_reference(upd2_ref_info)
    # Updated:
    assert model.prior.ref_pdf[0] == upd2_ref_info["a"]
    assert model.prior.ref_pdf[1] == upd2_ref_info["b"]
    # Unchanged:
    assert model.prior.ref_pdf[2] == val
    assert model.prior.ref_pdf[3] == upd1_ref_info["d"]
    assert model.prior.ref_pdf[4] == upd1_ref_info["e"]
    # Should be point-like now!
    assert model.prior.reference_is_pointlike
    # Let's update it -- back to one norm
    upd3_ref_info = {
        "a": [mean, std]}
    model.prior.set_reference(upd3_ref_info)
    # Updated:
    assert model.prior.ref_pdf[0].dist.name == "norm"
    assert model.prior.ref_pdf[0].mean() == mean
    assert model.prior.ref_pdf[0].std() == std
    # Unchanged:
    assert model.prior.ref_pdf[1] == upd2_ref_info["b"]
    assert model.prior.ref_pdf[2] == val
    assert model.prior.ref_pdf[3] == upd1_ref_info["d"]
    assert model.prior.ref_pdf[4] == upd1_ref_info["e"]
    # Should be non-point-like again
    assert not model.prior.reference_is_pointlike


# Tests MPI-awareness of the referece, when using it to get the initial point of mcmc
@pytest.mark.mpi
@mpi.sync_errors
def test_ref_mcmc_initial_point():
    val = 0.5
    info = {
        "params": {"a": {"prior": [0, 1 + mpi.size()], "ref": val + mpi.rank()}},
        "likelihood": {"one": None}}
    model = get_model(info)
    mcmc_sampler = get_sampler({"mcmc": None}, model)
    initial_point = mcmc_sampler.current_point.values[0]
    assert initial_point == val + mpi.rank()



"""
.. module:: cobaya.grid_tools.gridconfig

:Synopsis: Grid creator (Cobaya version)
:Author: Antony Lewis and Jesus Torrado
         (based on Antony Lewis' CosmoMC version of the same code)

"""

# Global
import os
import copy
import argparse

# Local
from cobaya.yaml import yaml_load_file, yaml_dump_file
from cobaya.conventions import Extension, packages_path_input
from cobaya.input import get_used_components, merge_info, update_info
from cobaya.install import install as install_reqs
from cobaya.tools import sort_cosmetic, warn_deprecation
from cobaya.grid_tools import batchjob
from cobaya.cosmo_input import create_input, get_best_covmat
from cobaya.parameterization import is_sampled_param


def getArgs(vals=None):
    parser = argparse.ArgumentParser(
        prog="cobaya grid-create",
        description='Initialize grid using settings file')
    parser.add_argument('batchPath', help=(
        'root directory containing/to contain the grid '
        '(e.g. ./PLA where directories base, base_xx etc are under ./PLA)'))
    parser.add_argument('settingName', nargs='?', help=(
        'python setting file (without .py) for making or updating a grid, '
        'usually found as python/settingName.py'))
    parser.add_argument('--read-only', action='store_true', help=(
        'option to configure an already-run existing grid'))
    # Arguments related to installation of requisites
    parser.add_argument('--install-reqs-at', help=(
        'install required code and data for the grid in the given folder.'))
    parser.add_argument("--install-reqs-force", action="store_true", default=False,
                        help="Force re-installation of apparently installed packages.")
    return parser.parse_args(vals)


def pathIsGrid(batchPath):
    return os.path.exists(batchjob.grid_cache_file(batchPath)) or os.path.exists(
        os.path.join(batchPath, 'config', 'config.ini'))


def make_grid_script():
    warn_deprecation()
    args = getArgs()
    args.interactive = True
    makeGrid(**args.__dict__)


def makeGrid(batchPath, settingName=None, settings=None, read_only=False,
             interactive=False, install_reqs_at=None, install_reqs_force=None):
    print("Generating grid...")
    batchPath = os.path.abspath(batchPath) + os.sep
    if not settings:
        if not settingName:
            raise NotImplementedError("Re-using previous batch is work in progress...")
        #            if not pathIsGrid(batchPath):
        #                raise Exception('Need to give name of setting file if batchPath/config '
        #                                'does not exist')
        #            read_only = True
        #            sys.path.insert(0, batchPath + 'config')
        #            settings = __import__(IniFile(batchPath + 'config/config.ini').params['setting_file'].replace('.py', ''))
        elif os.path.splitext(settingName)[-1].lower() in Extension.yamls:
            settings = yaml_load_file(settingName)
        else:
            raise NotImplementedError("Using a python script is work in progress...")
            # In this case, info-as-dict would be passed
            # settings = __import__(settingName, fromlist=['dummy'])
    batch = batchjob.BatchJob(batchPath)
    # batch.skip = settings.get("skip", False)
    batch.makeItems(settings, messages=not read_only)
    if read_only:
        for jobItem in [b for b in batch.jobItems]:
            if not jobItem.chainExists():
                batch.jobItems.remove(jobItem)
        batch.save()
        print('OK, configured grid with %u existing chains' % (len(batch.jobItems)))
        return batch
    else:
        batch.makeDirectories(setting_file=None)
        batch.save()
    infos = {}
    components_used = {}
    # Default info
    defaults = copy.deepcopy(settings)
    grid_definition = defaults.pop("grid")
    models_definitions = grid_definition["models"]
    datasets_definitions = grid_definition["datasets"]
    for jobItem in batch.items(wantSubItems=False):
        # Model info
        jobItem.makeChainPath()
        try:
            model_info = copy.deepcopy(models_definitions[jobItem.param_set] or {})
        except KeyError:
            raise ValueError("Model '%s' must be defined." % jobItem.param_set)
        model_info = merge_info(defaults, model_info)
        # Dataset info
        try:
            dataset_info = copy.deepcopy(datasets_definitions[jobItem.data_set.tag])
        except KeyError:
            raise ValueError("Data set '%s' must be defined." % jobItem.data_set.tag)
        # Combined info
        combined_info = merge_info(defaults, model_info, dataset_info)
        if "preset" in combined_info:
            preset = combined_info.pop("preset")
            combined_info = merge_info(create_input(**preset), combined_info)
        combined_info["output"] = jobItem.chainRoot
        # Requisites
        components_used = get_used_components(components_used, combined_info)
        if install_reqs_at:
            combined_info[packages_path_input] = os.path.abspath(install_reqs_at)
        # Save the info (we will write it after installation:
        # we need to install to add auto covmats
        if jobItem.param_set not in infos:
            infos[jobItem.param_set] = {}
        infos[jobItem.param_set][jobItem.data_set.tag] = combined_info
    # Installing requisites
    if install_reqs_at:
        print("Installing required code and data for the grid.")
        from cobaya.log import logger_setup
        logger_setup()
        install_reqs(components_used, path=install_reqs_at, force=install_reqs_force)
    print("Adding covmats (if necessary) and writing input files")
    for jobItem in batch.items(wantSubItems=False):
        info = infos[jobItem.param_set][jobItem.data_set.tag]
        # Covariance matrices
        # We try to find them now, instead of at run time, to check if correctly selected
        try:
            sampler = list(info["sampler"])[0]
        except KeyError:
            raise ValueError("No sampler has been chosen")
        if sampler == "mcmc" and info["sampler"][sampler].get("covmat", "auto"):
            packages_path = install_reqs_at or info.get(packages_path_input, None)
            if not packages_path:
                raise ValueError("Cannot assign automatic covariance matrices because no "
                                 "external packages path has been defined.")
            # Need updated info for covmats: includes renames
            updated_info = update_info(info)
            # Ideally, we use slow+sampled parameters to look for the covariance matrix
            # but since for that we'd need to initialise a model, we approximate that set
            # as theory+sampled
            from itertools import chain
            like_params = set(chain(*[
                list(like["params"])
                for like in updated_info["likelihood"].values()]))
            params_info = {p: v for p, v in updated_info["params"].items()
                           if is_sampled_param(v) and p not in like_params}
            best_covmat = get_best_covmat(os.path.abspath(packages_path),
                                          params_info, updated_info["likelihood"])
            info["sampler"][sampler]["covmat"] = os.path.join(
                best_covmat["folder"], best_covmat["name"])
        # Write the info for this job
        # Allow overwrite since often will want to regenerate grid with tweaks
        yaml_dump_file(jobItem.iniFile(), sort_cosmetic(info), error_if_exists=False)

        # Non-translated old code
        # if not start_at_bestfit:
        #     setMinimize(jobItem, ini)
        #     variant = '_minimize'
        #     ini.saveFile(jobItem.iniFile(variant))
        ## NOT IMPLEMENTED: start at best fit
        ##        ini.params['start_at_bestfit'] = start_at_bestfit
        # ---
        # for deffile in settings.defaults:
        #    ini.defaults.append(batch.commonPath + deffile)
        # if hasattr(settings, 'override_defaults'):
        #    ini.defaults = [batch.commonPath + deffile for deffile in settings.override_defaults] + ini.defaults
        # ---
        # # add ini files for importance sampling runs
        # for imp in jobItem.importanceJobs():
        #     if getattr(imp, 'importanceFilter', None): continue
        #     if batch.hasName(imp.name.replace('"post"', '')):
        #         raise Exception('importance sampling something you already have?')
        #     for minimize in (False, True):
        #         if minimize and not getattr(imp, 'want_minimize', True): continue
        #         ini = IniFile()
        #         updateIniParams(ini, imp.importanceSettings, batch.commonPath)
        #         if cosmomcAction == 0 and not minimize:
        #             for deffile in settings.importanceDefaults:
        #                 ini.defaults.append(batch.commonPath + deffile)
        #             ini.params['redo_outroot'] = imp.chainRoot
        #             ini.params['action'] = 1
        #         else:
        #             ini.params['file_root'] = imp.chainRoot
        #         if minimize:
        #             setMinimize(jobItem, ini)
        #             variant = '_minimize'
        #         else:
        #             variant = ''
        #         ini.defaults.append(jobItem.iniFile())
        #         ini.saveFile(imp.iniFile(variant))
        #         if cosmomcAction != 0: break

    if not interactive:
        return batch
    print('Done... to run do: cobaya-grid-run %s' % batchPath)
#    if not start_at_bestfit:
#        print('....... for best fits: python python/runbatch.py %s --minimize'%batchPath)
#    print('')
#    print('for importance sampled: python python/runbatch.py %s --importance'%batchPath)
#    print('for best-fit for importance sampled: '
#          'python python/runbatch.py %s --importance_minimize'%batchPath)



import fnmatch
import argparse

from cobaya.grid_tools import batchjob


class BatchArgs:
    def __init__(self, prog=None, desc='', importance=True, noBatchPath=False,
                 notExist=False, notall=False, converge=False,
                 plots=False, batchPathOptional=False):
        self.parser = argparse.ArgumentParser(prog=prog, description=desc)
        if not noBatchPath:
            if batchPathOptional:
                self.parser.add_argument('batchPath', nargs='?',
                                         help='directory containing the grid')
            else:
                self.parser.add_argument('batchPath',
                                         help='directory containing the grid')
        if converge:
            self.parser.add_argument('--converge', type=float, default=0,
                                     help='minimum R-1 convergence')
        self.importanceParameter = importance
        self.notExist = notExist
        self.notall = notall
        self.doplots = plots

    def parseForBatch(self, vals=None):
        if self.importanceParameter:
            self.parser.add_argument('--noimportance', action='store_true',
                                     help='original chains only, no importance sampled')
            self.parser.add_argument('--importance', nargs='*', default=None,
                                     help='data names for importance sampling '
                                          'runs to include')
            self.parser.add_argument('--importancetag', nargs='*', default=None,
                                     help='importance tags for importance '
                                          'sampling runs to include')

        self.parser.add_argument('--name', default=None, nargs='+',
                                 help='specific chain full name only '
                                      '(base_paramx_data1_data2)')
        self.parser.add_argument('--param', default=None, nargs='+',
                                 help='runs including specific parameter only (paramx)')
        self.parser.add_argument('--paramtag', default=None, nargs='+',
                                 help='runs with specific parameter tag only '
                                      '(base_paramx)')
        self.parser.add_argument('--data', nargs='+', default=None,
                                 help='runs including specific data only (data1)')
        self.parser.add_argument('--datatag', nargs='+', default=None,
                                 help='runs with specific data tag only (data1_data2)')
        self.parser.add_argument('--musthave-data', nargs='+', default=None,
                                 help='include only runs that include specific data '
                                      '(data1)')
        self.parser.add_argument('--skip-data', nargs='+', default=None,
                                 help='skip runs containing specific data (data1)')
        self.parser.add_argument('--skip-param', nargs='+', default=None,
                                 help='skip runs containing specific parameter (paramx)')
        self.parser.add_argument('--group', default=None, nargs='+',
                                 help='include only runs with given group names')
        self.parser.add_argument('--skip-group', default=None, nargs='+',
                                 help='exclude runs with given group names')

        if self.notExist:
            self.parser.add_argument('--notexist', action='store_true',
                                     help='only include chains that don\'t already '
                                          'exist on disk')
        if self.notall:
            self.parser.add_argument('--notall', type=int, default=None,
                                     help='only include chains where all N chains '
                                          'don\'t already exist on disk')
        if self.doplots:
            self.parser.add_argument('--paramNameFile', default='clik_latex.paramnames',
                                     help=".paramnames file for custom labels for "
                                          "parameters")
            self.parser.add_argument('--paramList', default=None,
                                     help=".paramnames file listing specific parameters "
                                          "to include (only)")
            self.parser.add_argument('--size_inch', type=float, default=None,
                                     help='output subplot size in inches')
            self.parser.add_argument('--nx', default=None, help='number of plots per row')
            self.parser.add_argument('--outputs', nargs='+', default=['pdf'],
                                     help='output file type (default: pdf)')

        args = self.parser.parse_args(vals)
        self.args = args
        if args.batchPath:
            self.batch = batchjob.readobject(args.batchPath)
            if self.batch is None:
                raise Exception('batchPath %s does not exist or is not '
                                'initialized with makeGrid.py' % args.batchPath)
            if self.doplots:
                import getdist.plots as plots
                from getdist import paramnames

                if args.paramList is not None:
                    args.paramList = paramnames.ParamNames(
                        args.paramList)
                g = plots.GetDistPlotter(chain_dir=self.batch.batchPath)
                if args.size_inch is not None:
                    g.settings.setWithSubplotSize(args.size_inch)
                return self.batch, self.args, g
            else:
                return self.batch, self.args
        else:
            return None, self.args

    def wantImportance(self, jobItem):
        return (self.args.importancetag is None or len(self.args.importancetag) == 0 or
                jobItem.importanceTag in self.args.importancetag) and \
               (self.args.importance is None or len(self.args.importance) == 0 or
                len([True for x in self.args.importance if
                     x in jobItem.data_set.importanceNames]))

    def jobItemWanted(self, jobItem):
        return not jobItem.isImportanceJob and (
                self.args.importance is None) or jobItem.isImportanceJob and \
               self.wantImportance(jobItem)

    def nameMatches(self, jobItem):
        if self.args.name is None:
            return True
        for pat in self.args.name:
            if fnmatch.fnmatch(jobItem.name, pat):
                return True
        return False

    def groupMatches(self, jobItem):
        return (self.args.group is None or jobItem.group in self.args.group) and (
                self.args.skip_group is None or jobItem.group not in self.args.skip_group)

    def dataMatches(self, jobItem):
        if self.args.musthave_data is not None and not jobItem.data_set.hasAll(
                self.args.musthave_data):
            return False
        if self.args.datatag is None:
            if self.args.skip_data is not None and jobItem.data_set.hasName(
                    self.args.skip_data):
                return False
            return self.args.data is None or jobItem.data_set.hasName(self.args.data)
        else:
            return jobItem.datatag in self.args.datatag

    def paramsMatch(self, jobItem):
        if self.args.paramtag is None:
            if self.args.param is None:
                return not self.args.skip_param or not jobItem.hasParam(
                    self.args.skip_param)
            for pat in self.args.param:
                if pat in jobItem.param_set:
                    return not self.args.skip_param or not jobItem.hasParam(
                        self.args.skip_param)
            return False
        else:
            return jobItem.paramtag in self.args.paramtag

    def filteredBatchItems(self, wantSubItems=True, chainExist=False):
        for jobItem in self.batch.items(wantImportance=not self.args.noimportance,
                                        wantSubItems=wantSubItems):
            if (not chainExist or jobItem.chainExists()) and (
                    self.jobItemWanted(jobItem) and self.nameMatches(jobItem) and
                    self.paramsMatch(jobItem) and self.dataMatches(jobItem)
                    and self.groupMatches(jobItem)):
                yield jobItem

    def sortedParamtagDict(self, chainExist=True):
        items = dict()
        for jobItem in self.filteredBatchItems():
            if not chainExist or jobItem.chainExists():
                if jobItem.paramtag not in items:
                    items[jobItem.paramtag] = []
                items[jobItem.paramtag].append(jobItem)
        return sorted(items.items())

    def filterForDataCompare(self, batch, datatags, getDistExists=False):
        items = []
        for tag, data in zip([self.batch.normalizeDataTag(data) for data in datatags],
                             datatags):
            items += [jobItem for jobItem in batch
                      if (jobItem.datatag == data or jobItem.normed_data == tag)
                      and (not getDistExists or jobItem.getDistExists())]
        return items



# Tries to evaluate the likelihood at LCDM's best fit of Planck 2015, with CAMB and CLASS
from copy import deepcopy

from .common_cosmo import body_of_test
from cobaya.cosmo_input import planck_precision

# Generating plots in Travis
try:
    import matplotlib

    matplotlib.use('agg')
except:
    pass

# Derived parameters not understood by CLASS
# https://wiki.cosmos.esa.int/planckpla2015/images/b/b9/Parameter_tag_definitions_2015.pdf
classy_unknown = ["zstar", "rstar", "thetastar", "DAstar", "zdrag",
                  "kd", "thetad", "zeq", "keq", "thetaeq", "thetarseq",
                  "DH", "Y_p"]

# Small chi2 difference with CLASS (total still <0.5)
classy_extra_tolerance = 0.45


# STANDARD ###############################################################################

def test_planck_2018_t_camb(packages_path, skip_not_installed, clik=False):
    best_fit = deepcopy(params_lowl_highTT_lensing)
    best_fit.pop("H0")
    info_likelihood = lik_info_lowl_highTT_lensing
    chi2 = chi2_lowl_highTT_lensing.copy()
    if clik:
        info_likelihood["planck_2018_lowl.TT_clik"] = \
            info_likelihood.pop("planck_2018_lowl.TT")
        chi2["planck_2018_lowl.TT_clik"] = chi2.pop("planck_2018_lowl.TT")
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    best_fit_derived = derived_lowl_highTT_lensing
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2,
                 best_fit_derived, skip_not_installed=skip_not_installed)


def test_planck_2018_t_clik_camb(packages_path, skip_not_installed):
    test_planck_2018_t_camb(packages_path, skip_not_installed, True)


def test_planck_2018_p_camb(packages_path, skip_not_installed, clik=False):
    best_fit = deepcopy(params_lowTE_highTTTEEE_lensingcmblikes)
    best_fit.pop("H0")
    info_likelihood = lik_info_lowTE_highTTTEEE_lensingcmblikes.copy()
    chi2 = chi2_lowTE_highTTTEEE_lensingcmblikes.copy()
    if clik:
        info_likelihood["planck_2018_lowl.EE_clik"] = \
            info_likelihood.pop("planck_2018_lowl.EE")
        chi2["planck_2018_lowl.EE_clik"] = chi2.pop("planck_2018_lowl.EE")
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    best_fit_derived = derived_lowTE_highTTTEEE_lensingcmblikes
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2,
                 best_fit_derived, skip_not_installed=skip_not_installed)


def test_planck_2018_p_clik_camb(packages_path, skip_not_installed):
    test_planck_2018_p_camb(packages_path, skip_not_installed, True)


# LITES ##################################################################################

def test_planck_2018_t_lite_camb(packages_path, skip_not_installed, native=False):
    best_fit = deepcopy(params_lowl_highTT_lite_lensing)
    best_fit.pop("H0")
    like_name = "planck_2018_highl_plik.TT_lite" + ("_native" if native else "")
    info_likelihood = {like_name: None}
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    chi2 = {like_name: chi2_planck_2018_plikHM_highTT_lite, "tolerance": 0.01}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2,
                 skip_not_installed=skip_not_installed)


def test_planck_2018_t_lite_native_camb(packages_path, skip_not_installed):
    test_planck_2018_t_lite_camb(packages_path, native=True,
                                 skip_not_installed=skip_not_installed)


def test_planck_2018_p_lite_camb(packages_path, skip_not_installed, native=False):
    best_fit = deepcopy(params_lowTE_highTTTEEE_lite_lensingcmblikes)
    best_fit.pop("H0")
    like_name = "planck_2018_highl_plik.TTTEEE_lite" + ("_native" if native else "")
    info_likelihood = {like_name: None}
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    chi2 = {like_name: chi2_planck_2018_plikHM_highTTTEEE_lite, "tolerance": 0.01}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2,
                 skip_not_installed=skip_not_installed)


def test_planck_2018_p_lite_native_camb(packages_path, skip_not_installed):
    test_planck_2018_p_lite_camb(packages_path, native=True,
                                 skip_not_installed=skip_not_installed)


# UNBINNED ###############################################################################

def test_planck_2018_t_unbinned_camb(packages_path, skip_not_installed):
    best_fit = deepcopy(params_lowl_highTT_lensing)
    best_fit.pop("H0")
    like_name = "planck_2018_highl_plik.TT_unbinned"
    info_likelihood = {like_name: None}
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    chi2 = {like_name: 8275.99, "tolerance": 0.03}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2,
                 skip_not_installed=skip_not_installed)


def test_planck_2018_p_unbinned_camb(packages_path, skip_not_installed):
    best_fit = deepcopy(params_lowTE_highTTTEEE_lensingcmblikes)
    best_fit.pop("H0")
    like_name = "planck_2018_highl_plik.TTTEEE_unbinned"
    info_likelihood = {like_name: None}
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    chi2 = {like_name: 24125.92, "tolerance": 0.01}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2,
                 skip_not_installed=skip_not_installed)


# CamSpec ################################################################################

def test_planck_2018_t_CamSpec_camb(packages_path, skip_not_installed):
    name = "planck_2018_highl_CamSpec.TT"
    info_likelihood = {name: None}
    chi2 = {name: 7060.04, 'tolerance': 0.2}
    best_fit = params_lowTE_highTTTEEE_lite_lensingcmblikes.copy()
    best_fit.update(
        {'aps100': 238.7887, 'aps143': 41.31762, 'aps217': 100.6226, 'acib217': 44.96003,
         'asz143': 5.886124,
         'psr': 0.5820399, 'cibr': 0.7912195, 'ncib': 0.0, 'cibrun': 0.0, 'xi': 0.1248677,
         'aksz': 1.153473,
         'dust100': 1.010905, 'dust143': 0.9905765, 'dust217': 0.9658913,
         'dust143x217': 0.9946434,
         'cal0': 0.9975484, 'cal2': 1.00139, 'calTE': 1.0, 'calEE': 1.0})
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2,
                 skip_not_installed=skip_not_installed)


def test_planck_2018_t_CamSpec2021_camb(packages_path, skip_not_installed):
    name = "planck_2018_highl_CamSpec2021.TT"
    info_likelihood = {name: None}
    chi2 = {name: 6358.67, 'tolerance': 0.2}
    best_fit = params_lowTE_highTTTEEE_lite_lensingcmblikes.copy()
    best_fit.pop("H0")
    best_fit.update(
        {'amp_143': 19.70529, 'amp_217': 15.25289, 'amp_143x217': 10.42223,
         'n_143': 0.94730, 'n_217': 1.05608, 'n_143x217': 0.96638})
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2,
                 skip_not_installed=skip_not_installed)


def test_planck_2018_p_CamSpec_camb(packages_path, skip_not_installed):
    name = "planck_2018_highl_CamSpec.TTTEEE"
    info_likelihood = {name: None}
    chi2 = {name: 11513.53, 'tolerance': 0.2}
    best_fit = params_lowTE_highTTTEEE_lite_lensingcmblikes.copy()
    best_fit.update(
        {'aps100': 238.7887, 'aps143': 41.31762, 'aps217': 100.6226, 'acib217': 44.96003,
         'asz143': 5.886124,
         'psr': 0.5820399, 'cibr': 0.7912195, 'ncib': 0.0, 'cibrun': 0.0, 'xi': 0.1248677,
         'aksz': 1.153473,
         'dust100': 1.010905, 'dust143': 0.9905765, 'dust217': 0.9658913,
         'dust143x217': 0.9946434,
         'cal0': 0.9975484, 'cal2': 1.00139, 'calTE': 1.0, 'calEE': 1.0})
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2,
                 skip_not_installed=skip_not_installed)


def test_planck_2018_p_CamSpec2021_camb(packages_path, skip_not_installed):
    name = "planck_2018_highl_CamSpec2021.TTTEEE"
    info_likelihood = {name: None}
    chi2 = {name: 10120.27, 'tolerance': 0.2}
    best_fit = params_lowTE_highTTTEEE_lite_lensingcmblikes.copy()
    best_fit.pop("H0")
    best_fit.update(
        {'amp_143': 17.49600, 'amp_217': 13.13760, 'amp_143x217': 8.26404,
         'n_143': 1.09456, 'n_217': 1.33554, 'n_143x217': 1.33658, 'calTE': 1.00048,
         'calEE': 1.00124})
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2,
                 skip_not_installed=skip_not_installed)


# CMB-marged lensing #####################################################################

def test_planck_2018_lcmbmarged_camb(packages_path, skip_not_installed):
    best_fit = params_lensing_cmbmarged
    info_likelihood = lik_info_lensing_cmbmarged
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    best_fit_derived = {}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_lensing_cmbmarged, best_fit_derived,
                 skip_not_installed=skip_not_installed)


# with CLASS #############################################################################

def test_planck_2018_t_classy(packages_path, skip_not_installed):
    best_fit = deepcopy(params_lowl_highTT_lensing)
    best_fit.pop("theta_MC_100")
    best_fit = params_lowl_highTT_lensing
    info_likelihood = lik_info_lowl_highTT_lensing
    info_theory = {"classy": {"extra_args": planck_precision["classy"]}}
    best_fit_derived = deepcopy(derived_lowl_highTT_lensing)
    for p in classy_unknown:
        best_fit_derived.pop(p, None)
    chi2_lowl_highTT_classy = deepcopy(chi2_lowl_highTT_lensing)
    chi2_lowl_highTT_classy["tolerance"] += classy_extra_tolerance
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_lowl_highTT_classy, best_fit_derived,
                 skip_not_installed=skip_not_installed)


def test_planck_2018_p_classy(packages_path, skip_not_installed):
    best_fit = deepcopy(params_lowTE_highTTTEEE_lensingcmblikes)
    best_fit.pop("theta_MC_100")
    info_likelihood = lik_info_lowTE_highTTTEEE_lensingcmblikes
    info_theory = {"classy": {"extra_args": planck_precision["classy"]}}
    best_fit_derived = deepcopy(derived_lowTE_highTTTEEE_lensingcmblikes)
    for p in classy_unknown:
        best_fit_derived.pop(p, None)
    chi2_lowl_highTT_classy = deepcopy(chi2_lowTE_highTTTEEE_lensingcmblikes)
    chi2_lowl_highTT_classy["tolerance"] += classy_extra_tolerance
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_lowl_highTT_classy, best_fit_derived,
                 skip_not_installed=skip_not_installed)


# Best fit temperature only ##############################################################

lik_info_lowl_highTT_lensing = {
    "planck_2018_lowl.TT": None, "planck_2018_highl_plik.TT": None,
    "planck_2018_lensing.clik": None}

chi2_lowl_highTT_lensing = {
    "planck_2018_lowl.TT": 22.92,
    "planck_2018_highl_plik.TT": 757.93,
    "planck_2018_lensing.clik": 9.11,
    "tolerance": 0.11}

chi2_planck_2018_plikHM_highTT_lite = 204.57

params_lowl_highTT_lite_lensing = {
    # Sampled
    "omegabh2": 0.02240,
    "omegach2": 0.1172,
    # only one of the next two is finally used!
    "H0": 68.45,  # will be ignored in the CAMB case
    "theta_MC_100": 1.04117,  # will be ignored in the CLASS case
    "tau": 0.0862,
    "logA": 3.100,
    "ns": 0.9733,
    # Planck likelihood
    "A_planck": 1.00008}

params_lowl_highTT_lensing = params_lowl_highTT_lite_lensing.copy()
params_lowl_highTT_lensing.update({
    "A_cib_217": 45.6,
    "xi_sz_cib": 0.73,
    "A_sz": 6.92,
    "ps_A_100_100": 246,
    "ps_A_143_143": 51.3,
    "ps_A_143_217": 54.7,
    "ps_A_217_217": 122.1,
    "ksz_norm": 0.01,
    "gal545_A_100": 8.87,
    "gal545_A_143": 10.79,
    "gal545_A_143_217": 19.8,
    "gal545_A_217": 95.4,
    "calib_100T": 0.99965,
    "calib_217T": 0.99822})

derived_lowl_highTT_lensing = {
    # param: [best_fit, sigma]
    "H0": [params_lowl_highTT_lensing["H0"], 1.2],
    "omegal": [0.7006, 0.015],
    "omegam": [0.2994, 0.016],
    "sigma8": [0.8271, 0.013],
    "zrei": [10.61, 2.2],
    "YHe": [0.245408, 0.0001],
    "Y_p": [0.246735, 0.0001],
    "DH": [2.5794e-5, 0.05e-5],
    "age": [13.7744, 0.047],
    "zstar": [1089.63, 0.52],
    "rstar": [145.13, 0.55],
    "thetastar": [1.04136, 0.00050],
    "DAstar": [13.9364, 0.050],
    "zdrag": [1059.818, 0.51],
    "rdrag": [147.79, 0.53],
    "kd": [0.14015, 0.00053],
    "thetad": [0.160844, 0.00029],
    "zeq": [3337, 57],
    "keq": [0.010184, 0.00017],
    "thetaeq": [0.8255, 0.011],
    "thetarseq": [0.4557, 0.0057]}

# Best fit polarization ##################################################################

lik_info_lowTE_highTTTEEE_lensingcmblikes = {
    "planck_2018_lowl.TT": None, "planck_2018_lowl.EE": None,
    "planck_2018_highl_plik.TTTEEE": None, "planck_2018_lensing.native": None}

chi2_lowTE_highTTTEEE_lensingcmblikes = {
    "planck_2018_lowl.TT": 23.25, "planck_2018_lowl.EE": 396.05,
    "planck_2018_highl_plik.TTTEEE": 2345.01, "planck_2018_lensing.native": 8.87,
    "tolerance": 0.11}

chi2_planck_2018_plikHM_highTTTEEE_lite = 584.64

params_lowTE_highTTTEEE_lite_lensingcmblikes = {
    # Sampled
    "omegabh2": 0.022383,
    "omegach2": 0.12011,
    # only one of the next two is finally used!
    "H0": 67.32,  # will be ignored in the CAMB case
    "theta_MC_100": 1.040909,  # will be ignored in the CLASS case
    "logA": 3.0448,
    "ns": 0.96605,
    "tau": 0.0543,
    # Planck likelihood
    "A_planck": 1.00044}

params_lowTE_highTTTEEE_lensingcmblikes = params_lowTE_highTTTEEE_lite_lensingcmblikes.copy()
params_lowTE_highTTTEEE_lensingcmblikes.update(
    {"A_cib_217": 46.1,
     "xi_sz_cib": 0.66,
     "A_sz": 7.08,
     "ps_A_100_100": 248.2,
     "ps_A_143_143": 50.7,
     "ps_A_143_217": 53.3,
     "ps_A_217_217": 121.9,
     "ksz_norm": 0.00,
     "gal545_A_100": 8.80,
     "gal545_A_143": 11.01,
     "gal545_A_143_217": 20.16,
     "gal545_A_217": 95.5,
     "galf_TE_A_100": 0.1138,
     "galf_TE_A_100_143": 0.1346,
     "galf_TE_A_100_217": 0.479,
     "galf_TE_A_143": 0.225,
     "galf_TE_A_143_217": 0.665,
     "galf_TE_A_217": 2.082,
     "calib_100T": 0.99974,
     "calib_217T": 0.99819})

derived_lowTE_highTTTEEE_lensingcmblikes = {
    # param: [best_fit, sigma]
    "H0": [params_lowTE_highTTTEEE_lensingcmblikes["H0"], 0.54],
    "omegal": [0.6842, 0.0073],
    "omegam": [0.3158, 0.0073],
    "sigma8": [0.8120, 0.0060],
    "zrei": [7.68, 7.67],
    "YHe": [0.245401, 0.000057],
    "Y_p": [0.246727, 0.000058],
    "DH": [2.5831e-5, 0.027e-5],
    "age": [13.7971, 0.023],
    "zstar": [1089.914, 0.25],
    "rstar": [144.394, 0.26],
    "thetastar": [1.041085, 0.00031],
    "DAstar": [13.8696, 0.025],
    "zdrag": [1059.971, 0.30],
    "rdrag": [147.049, 0.26],
    "kd": [0.140922, 0.00030],
    "thetad": [0.160734, 0.00017],
    "zeq": [3405.1, 26],
    "keq": [0.010393, 0.000081],
    "thetaeq": [0.81281, 0.0050],
    "thetarseq": [0.44912, 0.0026]}

# Best fit CMB-marged lensing ############################################################

lik_info_lensing_cmbmarged = {"planck_2018_lensing.CMBMarged": None}

chi2_lensing_cmbmarged = {
    "planck_2018_lensing.CMBMarged": 7.51, "tolerance": 0.11}

params_lensing_cmbmarged = {
    "omegabh2": 2.2219050E-02,
    "omegach2": 1.1726920E-01,
    "theta_MC_100": 1.1180650E+00,
    "tau": 0.055,
    "logA": 3.2528000E+00,
    "ns": 9.6135180E-01}



"""
Testing and automatic generation of basic cosmological examples in the docs.
"""

import os
import sys

from cobaya.yaml import yaml_dump, yaml_load
from cobaya.input import is_equal_info
from cobaya.cosmo_input import create_input
from cobaya.tools import sort_cosmetic
from .test_docs_example_quickstart import docs_folder

path = os.path.join(docs_folder, "src_examples", "cosmo_basic")
file_pre = "basic_"
preset_pre = "planck_2018_"


def test_cosmo_docs_basic():
    flag = True
    for theo in ["camb", "classy"]:
        info_new = create_input(preset=preset_pre + theo)
        info_yaml_new = yaml_dump(sort_cosmetic(info_new))
        file_path = os.path.join(path, file_pre + theo + ".yaml")
        with open(file_path) as docs_file:
            info_yaml_docs = "".join(docs_file.readlines())
        info_docs = yaml_load(info_yaml_docs)
        if not is_equal_info(info_new, info_docs, strict=True, print_not_log=True):
            with open(file_path, "w") as docs_file:
                docs_file.write(info_yaml_new)
            flag = False
            print("OLD:\n%s" % info_yaml_docs)
            print("----------------------------------------")
            print("NEW:\n%s" % info_yaml_new)
            sys.stdout.flush()
    assert flag, ("Differences in example input file. "
                  "Files have been re-generated; check out your git diff.")



import logging
import os
import numpy as np
import pytest
from getdist.mcsamples import MCSamplesFromCobaya, loadMCSamples
from cobaya import mpi, run, Theory, InputDict, PostDict, LoggedError
from cobaya.conventions import Extension
from cobaya.component import ComponentNotFoundError
from cobaya.tools import deepcopy_where_possible
from cobaya.cosmo_input.convert_cosmomc import cosmomc_root_to_cobaya_info_dict
from cobaya.log import NoLogging
from .common import process_packages_path

pytestmark = pytest.mark.mpi


def likelihood(_self):
    return -(_self.provider.get_param('sigma8') - 0.7) ** 2 / 0.1 ** 2


def likelihood2(sigma8, joint):
    return -(sigma8 - 0.75) ** 2 / 0.07 ** 2 + joint * 0


class ATheory(Theory):
    params = {'As': None, 'As100': {'derived': True}}

    def calculate(self, state, want_derived=True, **params_values_dict):
        state['derived']['As100'] = params_values_dict['As'] * 100


class BTheory(Theory):
    params = {'As100': None, 'As1000': {'derived': True}}

    def calculate(self, state, want_derived=True, **params_values_dict):
        state['derived']['As1000'] = params_values_dict['As100'] * 10


class CTheory(Theory):
    params = {'AsX': None, 'As1000': {'derived': True}}


info: InputDict = {"params": {
    "ombh2": 0.0245,
    "H0": 70,
    "ns": 0.965,
    "logA": {'prior': [1, 4], 'latex': r'\log(10^{10} A_\mathrm{s})', "proposal": 0.1},
    "As": {'value': lambda logA: 1e-10 * np.exp(logA), 'latex': r'A_\mathrm{s}'},
    "omegab": {'value': lambda ombh2, H0: ombh2 / (H0 / 100) ** 2,
               'latex': r'\Omega_\mathrm{b}'},
    "omegam": {'value': 0.3, 'latex': r'\Omega_\mathrm{m}', 'drop': True},
    "omch2": {'value': lambda omegam, omegab, mnu, H0: (omegam - omegab) * (
            H0 / 100) ** 2 - (mnu * (3.044 / 3) ** 0.75) / 94.0708,
              'latex': r'\Omega_\mathrm{c} h^2'},
    "tau": 0.05,
    "mnu": 0.0,
    "sigma8": None,
    "sigma82": {"derived": "lambda sigma8, omegam: sigma8 ** 2 * omegam"},
    "joint": {"derived": "lambda sigma8, omegam, As100: sigma8 ** 2 * omegam*As100"}},
    "likelihood": {
        "test_likelihood": {'external': likelihood,
                            'type': 'type1',
                            'requires': ['As', 'sigma8']}},
    'theory': {'test_theory': ATheory,
               'camb': {'stop_at_error': True,
                        'extra_args': {'num_massive_neutrinos': 1,
                                       'halofit_version': 'mead'}}},
    'sampler': {'mcmc': {'Rminus1_stop': 0.3}}
}


def test_cosmo_run_not_found():
    with NoLogging(logging.ERROR):
        inf = deepcopy_where_possible(info)
        inf["likelihood"]["H0.perfect"] = None
        with pytest.raises(ComponentNotFoundError):
            run(inf)
        inf = deepcopy_where_possible(info)
        inf["likelihood"]["none"] = None
        with pytest.raises(ComponentNotFoundError):
            run(inf)
        inf = deepcopy_where_possible(info)
        inf["likelihood"]["pandas.plotting.PlotAccessor"] = None
        with pytest.raises(LoggedError) as e:
            run(inf)
        assert "Failed to get defaults for component" in str(e)


@mpi.sync_errors
def test_cosmo_run_resume_post(tmpdir, packages_path=None):
    # only vary As, so fast chain. Chain does not need to converge (tested elsewhere).
    info['output'] = os.path.join(tmpdir, 'testchain')
    if packages_path:
        info["packages_path"] = process_packages_path(packages_path)
    run(info, force=True)
    # note that continuing from files leads to text-file precision at read in, so a mix of
    # precision in the output SampleCollection returned from run
    run(info, resume=True, override={'sampler': {'mcmc': {'Rminus1_stop': 0.2}}})
    updated_info, sampler = run(info['output'] + '.updated' + Extension.dill,
                                resume=True,
                                override={'sampler': {'mcmc': {'Rminus1_stop': 0.05}}})
    results = mpi.allgather(sampler.products()["sample"])
    samp = MCSamplesFromCobaya(updated_info, results, ignore_rows=0.2)
    assert np.isclose(samp.mean('As100'), 100 * samp.mean('As'))

    # post-processing
    info_post: PostDict = {'add': {'params': {'h': None},
                                   "likelihood": {"test_likelihood2": likelihood2}},
                           'remove': {'likelihood': ["test_likelihood"]},
                           'suffix': 'testpost',
                           'skip': 0.2, 'thin': 4
                           }

    output_info, products = run(updated_info, override={'post': info_post}, force=True)
    results2 = mpi.allgather(products["sample"])
    samp2 = MCSamplesFromCobaya(output_info, results2)
    samp_test = samp.copy()
    samp_test.weighted_thin(4)
    sigma8 = samp_test.getParams().sigma8
    samp_test.reweightAddingLogLikes(-(sigma8 - 0.7) ** 2 / 0.1 ** 2
                                     + (sigma8 - 0.75) ** 2 / 0.07 ** 2)
    assert np.isclose(samp_test.mean('sigma8'), samp2.mean('sigma8'))

    # from getdist-format chain files
    root = os.path.join(tmpdir, 'getdist_format')
    if mpi.is_main_process():
        samp.saveChainsAsText(root)
    mpi.sync_processes()

    from_txt = dict(updated_info, output=root)
    post_from_text = dict(info_post, skip=0)  # getdist already skipped
    output_info, products = run(from_txt, override={'post': post_from_text}, force=True)
    samp_getdist = MCSamplesFromCobaya(output_info, mpi.allgather(products["sample"]))
    assert not products["stats"]["points_removed"]
    assert samp2.numrows == samp_getdist.numrows
    assert np.isclose(samp2.mean('sigma8'), samp_getdist.mean('sigma8'))

    # again with inferred-inputs for params
    info_conv = cosmomc_root_to_cobaya_info_dict(root)
    # have to manually add consistent likelihoods if re-computing
    info_conv['likelihood'] = info['likelihood']
    info_conv['theory'] = info['theory']
    post_from_text = dict(info_post, skip=0, suffix='getdist2')  # getdist already skipped
    output_info, products = run(info_conv, override={'post': post_from_text},
                                output=False)
    samp_getdist2 = MCSamplesFromCobaya(output_info, mpi.allgather(products["sample"]))
    assert np.isclose(samp2.mean('sigma8'), samp_getdist2.mean('sigma8'))

    # from save info, no output
    info_post['output'] = None
    output_info, products = run({'output': info['output'], 'post': info_post}, force=True)
    results3 = mpi.allgather(products["sample"])
    samp3 = MCSamplesFromCobaya(output_info, results3)
    assert np.isclose(samp3.mean("sigma8"), samp2.mean("sigma8"))
    assert np.isclose(samp3.mean("joint"), samp2.mean("joint"))
    samps4 = loadMCSamples(info['output'] + '.post.testpost')
    assert np.isclose(samp3.mean("joint"), samps4.mean("joint"))

    # test recover original answer swapping likelihoods back
    info_revert = {'add': {'likelihood': info['likelihood']},
                   'remove': {'likelihood': ["test_likelihood2"]},
                   'suffix': 'revert',
                   'skip': 0, 'thin': 1,
                   'output': None
                   }
    output_info, products = run({'output': info['output'] + '.post.testpost',
                                 'post': info_revert}, force=True)
    results_revert = mpi.allgather(products["sample"])
    samp_revert = MCSamplesFromCobaya(output_info, results_revert)

    samp_thin = MCSamplesFromCobaya(updated_info, results, ignore_rows=0.2)
    samp_thin.weighted_thin(4)
    assert samp_thin.numrows == samp_revert.numrows + products["stats"]["points_removed"]
    if not products["stats"]["points_removed"]:
        assert np.isclose(samp_revert.mean("sigma8"), samp_thin.mean("sigma8"))
    else:
        assert abs(samp_revert.mean("sigma8") - samp_thin.mean("sigma8")) < 0.01
    assert not products["stats"]["points_removed"]

    # no remove
    info_post = {
        'add': {'params': {'h': None}, "likelihood": {"test_likelihood2": likelihood2}},
        'suffix': 'test2', 'skip': 0.2, 'thin': 4}
    output_info, products = run(updated_info, override={'post': info_post}, force=True)
    results2 = mpi.allgather(products["sample"])
    samp2 = MCSamplesFromCobaya(output_info, results2)
    assert "chi2__type1" in samp2.paramNames.list()
    # check what has been saved to disk is consistent
    samps4 = loadMCSamples(updated_info['output'] + '.post.test2')
    assert samp2.paramNames.list() == samps4.paramNames.list()
    assert np.isclose(samp2.mean("sigma8"), samps4.mean("sigma8"))

    # adding new theory derived
    info_post['add']['theory'] = {'new_param_theory': BTheory}
    output_info, products = run(updated_info, override={'post': info_post}, output=False)
    results3 = mpi.allgather(products["sample"])
    samp3 = MCSamplesFromCobaya(output_info, results3)
    assert np.isclose(samp3.mean("sigma8"), samp2.mean("sigma8"))
    assert np.isclose(samp3.mean("As1000"), samp2.mean("As") * 1000)

    info_post['add']['theory'] = {'new_param_theory': CTheory}
    with pytest.raises(LoggedError) as e, NoLogging(logging.ERROR):
        run(updated_info, override={'post': info_post}, output=False)
    assert 'Parameter AsX no known value' in str(e)



"""
.. module:: containers

:Synopsis: Functions and scripts to manage container images
:Author: Jesus Torrado

"""

# Global
import os
from io import StringIO
from tempfile import NamedTemporaryFile
from subprocess import Popen, PIPE
import uuid
import argparse
from textwrap import dedent
from requests import head

# Local
from cobaya.log import logger_setup, LoggedError, get_logger
from cobaya.input import get_used_components, load_input
from cobaya.yaml import yaml_dump
from cobaya.install import install
from cobaya.conventions import products_path, packages_path_env, packages_path_arg, \
    code_path, data_path, packages_path_input
from cobaya.tools import warn_deprecation

log = get_logger(__name__)

_requirements_file = "requirements.yaml"
_help_file = "readme.md"

requirements_file_path = os.path.join(packages_path_input, _requirements_file)
help_file_path = os.path.join(packages_path_input, _help_file)

base_recipe = r"""
# OS -------------------------------------------------------------------------
FROM ubuntu:xenial
# POST -----------------------------------------------------------------------
RUN sed -i 's/# \(.*multiverse$\)/\1/g' /etc/apt/sources.list && \
    apt-get update && \
    apt-get -y upgrade && \
    apt-get install -y \
      autoconf automake make gcc-6-base nano \
      libopenblas-base liblapack3 liblapack-dev libcfitsio-dev \
      python python-pip git wget
# Python requisites -- LC_ALL=C: Necessary just for pip <= 8.1.2 (Xenial version)
ENV LC_ALL C
RUN python -m pip install --upgrade pip
RUN python -m pip install pytest-forked matplotlib cython astropy --upgrade
# Prepare environment and tree for external packages -------------------------
ENV LD_LIBRARY_PATH $LD_LIBRARY_PATH:/usr/local/lib
ENV CONTAINED TRUE
ENV %s %s
ENV COBAYA_PRODUCTS %s
RUN mkdir $%s && \
    mkdir $COBAYA_PRODUCTS
# COBAYA  --------------------------------------------------------------------
# getdist fork (it will be an automatic requisite in the future)
RUN cd $%s && git clone https://github.com/JesusTorrado/cobaya.git && \
    cd $%s/cobaya && python -m pip install -e .
""" % (packages_path_env, packages_path_input, products_path,
       packages_path_env, packages_path_env, packages_path_env)

MPI_URL = {
    "mpich": "https://www.mpich.org/static/downloads/_VER_/mpich-_VER_.tar.gz",
    "openmpi": ("https://www.open-mpi.org/software/ompi/v_VER_/downloads/"
                "openmpi-_VER__DOT_SUB_.tar.gz")}

MPI_versions = {"mpich": {"3.2": None},
                "openmpi": {"2.1": ["1"]}}  # , "2"]}}

MPI_recipe = {
    "mpich": """
    # NERSC: must be MPICH >= 3.2
    # https://www.nersc.gov/users/software/using-shifter-and-docker/using-shifter-at-nersc/
    RUN cd /tmp && wget _URL_ && tar xvzf mpich-_VER_.tar.gz && cd /tmp/mpich-_VER_ && \
      ./configure --prefix=/usr/local && make -j4 && make install && make clean && cd .. \
      rm -rf /tmp/mpich-* """,
    "openmpi": """
    # HPC: must be OpenMPI >= 2.1
    RUN cd /tmp && wget _URL_ && gunzip -c openmpi-_VER__DOT_SUB_.tar.gz | tar xf - && \
      cd /tmp/openmpi-_VER__DOT_SUB_ && ./configure --prefix=/usr/local && make -j4 && \
      make install && make clean && cd .. && rm -rf /tmp/openmpi-_VER__DOT_SUB_ """}

MPI_epilogue = """&& export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/usr/local/lib/ && \
                  ldconfig && python -m pip install mpi4py --no-binary :all:"""


def image_help(engine):
    e = engine.lower()
    assert e in ("singularity", "docker"), e + " not valid."
    mount_data = {"docker": "-v [/cluster/path/to/data]:/packages/data:rw",
                  "singularity": "--bind [/cluster/path/to/data]:/packages/data"}
    mount_products = {"docker": "-v [/cluster/path/to/products]:/products:rw",
                      "singularity": "--bind [/cluster/path/to/producs]:/products"}
    mount_tmp = {"docker": "-v /tmp:/products:rw",
                 "singularity": "--bind /tmp:/products"}
    pre_prepare = {"docker": " ".join(["?????", mount_data[e]]),
                   "singularity": " ".join(
                       ["singularity exec", mount_data[e], "[image_file]"])}
    pre_run = {"docker": " ".join(["?????", mount_data[e], mount_products[e]]),
               "singularity": " ".join(
                   [mount_data[e], mount_products[e], "[image_file]"])}
    pre_shell = {"docker": " ".join(["?????", mount_data[e], mount_tmp[e]]),
                 "singularity": " ".join(
                     ["singularity shell", mount_data[e], mount_tmp[e], "[image_file]"])}
    return dedent("""
        This is a %s image for Cobaya.

        To check the packages installed in the container, take a look at '%s'.

        Make sure that you have created a 'data' and a 'products' folder in your cluster.

        To prepare the data needed for the container, while in the cluster, run:

            $ %s cobaya-prepare-data  # --force

        To run a sample with an input file 'somename.yaml', send to you cluster scheduler:

            $ mpi[run|exec] [mpi options] %s cobaya-run somename.yaml

        To open a terminal in the container, for testing purposes do:

            $ %s

        Have fun!
        """ % (engine.title(), requirements_file_path, pre_prepare[engine.lower()],
               pre_run[engine.lower()], pre_shell[engine.lower()]))


def get_docker_client():
    try:
        import docker
    except ImportError:
        raise LoggedError(log,
                          "The Python Docker interface not installed: "
                          "do 'python -m pip install docker'.")
    return docker.from_env(version="auto")


def create_base_image(mpi=None, version=None, sub=None):
    """
    `mpi` : MPI library to compile: OpenMPI or MPICH.
    `ver` : major.minor, as a string.
    `sub` : (optional) the subversion of the major.minor version
    """
    if version is None:
        raise LoggedError(log, "Needs a major.minor version!")
    sub = sub or ""
    if sub:
        sub = "." + sub
    try:
        tag = "cobaya/base_%s_%s:latest" % (mpi.lower(), version + sub)
    except KeyError:
        raise LoggedError(log, "MPI library '%s' not recognized.", mpi)
    URL = MPI_URL[mpi.lower()].replace("_VER_", str(version)).replace("_DOT_SUB_", sub)
    if head(URL).status_code >= 400:
        raise LoggedError(log, "Failed to download %s %s: couldn't reach URL: '%s'",
                          mpi.lower(), version + sub, URL)
    log.info("Creating base image %s v%s...", mpi.lower(), version + sub)
    this_MPI_recipe = dedent(MPI_recipe[mpi.lower()].replace("_VER_", version)
                             .replace("_DOT_SUB_", sub).replace("_URL_", URL))
    dc = get_docker_client()
    with StringIO(base_recipe + this_MPI_recipe + MPI_epilogue) as stream:
        dc.images.build(fileobj=stream, tag=tag, nocache=True)
    log.info("Base image '%s' created!", tag)


def create_all_base_images():
    log.info("Creating all base images. This will take a while.")
    for mpi in MPI_recipe:
        for version, subs in MPI_versions[mpi.lower()].items():
            for sub in (subs or [None]):
                create_base_image(mpi, version, sub)
    log.info("All base images created. Don't forget to *PUSH*!")


def create_docker_image(filenames, MPI_version=None):
    log.info("Creating Docker image...")
    if not MPI_version:
        MPI_version = "3.2"
        # log.warning("You have not specified an MPICH version. "
        #          "It is strongly encouraged to request the one installed in your cluster,"
        #          " using '--mpi-version X.Y'. Defaulting to MPICH v%s.", MPI_version)
    dc = get_docker_client()
    components = yaml_dump(
        get_used_components(*[load_input(f) for f in filenames])).strip()
    echos_reqs = "RUN " + " && \\ \n    ".join(
        [r'echo "%s" >> %s' % (block, requirements_file_path)
         for block in components.split("\n")])
    echos_help = "RUN " + " && \\ \n    ".join(
        [r'echo "%s" >> %s' % (line, help_file_path)
         for line in image_help("docker").split("\n")])
    recipe = r"""
    FROM cobaya/base_mpich_%s:latest
    %s
    RUN cobaya-install %s --%s %s --just-code --force ### NEEDS PYTHON UPDATE! --no-progress-bars
    %s
    CMD ["cat", "%s"]
    """ % (MPI_version, echos_reqs, requirements_file_path, packages_path_arg,
           packages_path_input, echos_help, help_file_path)
    image_name = "cobaya:" + uuid.uuid4().hex[:6]
    with StringIO(recipe) as stream:
        dc.images.build(fileobj=stream, tag=image_name)
    log.info("Docker image '%s' created! "
             "Do 'docker save %s | gzip > some_name.tar.gz'"
             "to save it to the current folder.", image_name, image_name)


def create_singularity_image(filenames, MPI_version=None):
    log.info("Creating Singularity image...")
    if not MPI_version:
        MPI_version = "2.1.1"
        # log.warning("You have not specified an OpenMPI version. "
        #          "It is strongly encouraged to request the one installed in your cluster,"
        #          " using '--mpi-version X.Y.Z'. Defaulting to OpenMPI v%s.", MPI_version)
    components = yaml_dump(
        get_used_components(*[load_input(f) for f in filenames])).strip()
    echos_reqs = "\n    " + "\n    ".join(
        [""] + ['echo "%s" >> %s' % (block, requirements_file_path)
                for block in components.split("\n")])
    recipe = (
            dedent("""
        Bootstrap: docker
        From: cobaya/base_openmpi_%s:latest\n
        %%post\n""" % MPI_version) +
            dedent(echos_reqs) + dedent("""
        export CONTAINED=TRUE
        cobaya-install %s --%s %s --just-code --force ### --no-progress-bars
        mkdir $COBAYA_PRODUCTS

        %%help

        %s
        """ % (requirements_file_path,
               # TODO: this looks wrong?
               packages_path_input,
               os.path.join(packages_path_arg, packages_path_input, data_path),
               "\n        ".join(image_help("singularity").split("\n")[1:]))))
    with NamedTemporaryFile(delete=False) as recipe_file:
        recipe_file.write(recipe.encode('utf-8'))
        recipe_file_name = recipe_file.name
    image_name = "cobaya_" + uuid.uuid4().hex[:6] + ".simg"
    process_build = Popen(["singularity", "build", image_name, recipe_file_name],
                          stdout=PIPE, stderr=PIPE)
    out, err = process_build.communicate()
    if process_build.returncode:
        log.info(out)
        log.info(err)
        raise LoggedError(log, "Image creation failed! See error message above.")
    log.info("Singularity image '%s' created!", image_name)


# Command-line scripts ###################################################################

def create_image_script():
    warn_deprecation()
    logger_setup()
    parser = argparse.ArgumentParser(
        prog="cobaya create-image",
        description=(
            "Cobaya's tool for preparing Docker (for Shifter) and Singularity images."))
    parser.add_argument("files", action="store", nargs="+", metavar="input_file.yaml",
                        help="One or more input files.")
    parser.add_argument("-v", "--mpi-version", action="store", default=None,
                        metavar="X.Y(.Z)", dest="version", help="Version of the MPI lib.")
    group_type = parser.add_mutually_exclusive_group(required=True)
    group_type.add_argument("-d", "--docker", action="store_const", const="docker",
                            help="Create a Docker image (for Shifter).", dest="type")
    group_type.add_argument("-s", "--singularity", action="store_const", dest="type",
                            const="singularity", help="Create a Singularity image.")
    arguments = parser.parse_args()
    if arguments.type == "docker":
        create_docker_image(arguments.files, MPI_version=arguments.version)
    elif arguments.type == "singularity":
        create_singularity_image(arguments.files, MPI_version=arguments.version)


def prepare_data_script():
    warn_deprecation()
    logger_setup()
    if "CONTAINED" not in os.environ:
        raise LoggedError(log, "This command should only be run within a container. "
                               "Run 'cobaya-install' instead.")
    parser = argparse.ArgumentParser(
        prog="cobaya prepare-data",
        description="Cobaya's installation tool for the data needed by a container.")
    parser.add_argument("-f", "--force", action="store_true", default=False,
                        help="Force re-installation of apparently installed packages.")
    arguments = parser.parse_args()
    try:
        info = load_input(requirements_file_path)
    except IOError:
        raise LoggedError(log, "Cannot find the requirements file. "
                               "This should not be happening.")
    install(info, path=packages_path_input, force=arguments.force,
            **{code_path: False, data_path: True})



from cobaya.likelihoods.base_classes import SN


class jla_lite(SN):
    r"""
    Likelihood (marginalized over nuisance parameters) of the JLA type Ia supernova sample
    \cite{Betoule:2014frx}, based on observations obtained by the SDSS-II and SNLS
    collaborations.
    """
    pass



"""
.. module:: mpi

:Synopsis: Manages MPI parallelization transparently
:Author: Jesus Torrado

"""

import os
import sys
import functools
from typing import List, Iterable, Union, Callable, Any, Optional
import numpy as np
import time
from enum import IntEnum

default_error_timeout_seconds = 5

# Vars to keep track of MPI parameters
_mpi: Any = None if os.environ.get('COBAYA_NOMPI', False) else -1
_mpi_size = -1
_mpi_comm: Any = -1
_mpi_rank: Optional[int] = -1


def set_mpi_disabled(disabled=True):
    """
    Disable MPI, e.g. for use on cluster head nodes where mpi4py may be installed
    but no MPI functions will work.
    """
    global _mpi, _mpi_size, _mpi_rank, _mpi_comm
    if disabled:
        _mpi = None
        _mpi_size = 0
        _mpi_comm = None
        _mpi_rank = None
    else:
        _mpi = -1
        _mpi_size = -1
        _mpi_comm = -1
        _mpi_rank = -1


def is_disabled():
    return _mpi is None


# noinspection PyUnresolvedReferences
def get_mpi():
    """
    Import and returns the MPI object, or None if not running with MPI.

    Can be used as a boolean test if MPI is present.
    """
    global _mpi
    if _mpi == -1:
        try:
            from mpi4py import MPI
            _mpi = MPI
        except ImportError:
            _mpi = None
        else:
            if more_than_one_process():
                try:
                    import dill
                except ImportError:
                    pass
                else:
                    _mpi.pickle.__init__(dill.dumps, dill.loads)

    return _mpi


def get_mpi_size():
    """
    Returns the number of MPI processes that have been invoked,
    or 0 if not running with MPI.
    """
    global _mpi_size
    if _mpi_size == -1:
        _mpi_size = getattr(get_mpi_comm(), "Get_size", lambda: 0)()
    return _mpi_size


def get_mpi_comm():
    """
    Returns the MPI communicator, or `None` if not running with MPI.
    """
    global _mpi_comm
    if _mpi_comm == -1:
        _mpi_comm = getattr(get_mpi(), "COMM_WORLD", None)
    return _mpi_comm


def get_mpi_rank():
    """
    Returns the rank of the current MPI process:
        * None: not running with MPI
        * Z>=0: process rank, when running with MPI

    Can be used as a boolean that returns `False` for both the root process,
    if running with MPI, or always for a single process; thus, everything under
    `if not(get_mpi_rank()):` is run only *once*.
    """
    global _mpi_rank
    if _mpi_rank == -1:
        _mpi_rank = getattr(get_mpi_comm(), "Get_rank", lambda: None)()
    return _mpi_rank


# Aliases for simpler use
def is_main_process():
    """
    Returns true if primary process or MPI not available.
    """
    return not bool(get_mpi_rank())


def more_than_one_process():
    return get_mpi_size() > 1


def check_errors():
    if process_state:
        process_state.check_error()


def sync_processes():
    if get_mpi_size() > 1:
        check_errors()
        get_mpi_comm().barrier()


def share_mpi(data=None, root=0):
    if get_mpi_size() > 1:
        return get_mpi_comm().bcast(data, root=root)
    else:
        return data


share = share_mpi


def scatter(data=None, root=0):
    if get_mpi_size() > 1:
        return get_mpi_comm().scatter(data, root=root)
    else:
        return data[0]


def size() -> int:
    return get_mpi_size() or 1


def rank() -> int:
    return get_mpi_rank() or 0


def gather(data, root=0) -> list:
    comm = get_mpi_comm()
    if comm and more_than_one_process():
        return comm.gather(data, root=root) or []
    else:
        return [data]


def allgather(data) -> list:
    if get_mpi_size() > 1:
        return get_mpi_comm().allgather(data)
    else:
        return [data]


def zip_gather(list_of_data, root=0) -> Iterable[tuple]:
    """
    Takes a list of items and returns an iterable of lists of items from each process
    e.g. for root node
    [(a_1, a_2),(b_1,b_2),...] = zip_gather([a,b,...])
    """
    if get_mpi_size() > 1:
        return zip(*(get_mpi_comm().gather(list_of_data, root=root) or [list_of_data]))
    else:
        return ((item,) for item in list_of_data)


def array_gather(list_of_data, root=0) -> List[np.ndarray]:
    return [np.array(i) for i in zip_gather(list_of_data, root=root)]


# set if being run from pytest
capture_manager: Any = None


def abort_if_mpi():
    """Closes all MPI process, if more than one present."""
    if get_mpi_size() > 1:
        if capture_manager:
            capture_manager.stop_global_capturing()
        get_mpi_comm().Abort(1)


_other_process_msg = "Another process failed - exiting."


class OtherProcessError(Exception):
    pass


def wait_for_request(req, time_out_seconds=default_error_timeout_seconds, interval=0.01):
    time_start = time.time()
    while not req.Test():
        time.sleep(interval)
        if time.time() - time_start > time_out_seconds:
            return False
    return True


def time_out_barrier(time_out_seconds=default_error_timeout_seconds):
    if more_than_one_process():
        return wait_for_request(_mpi_comm.Ibarrier(), time_out_seconds)
    return True


# decorators to generalize functions/methods for mpi sharing

def root_only(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if is_main_process():
            return func(*args, **kwargs)

    return wrapper


def more_than_one(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if more_than_one_process():
            return func(*args, **kwargs)

    return wrapper


def from_root(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if is_main_process():
            try:
                result = func(*args, **kwargs)
            except Exception:
                share_mpi()
                raise
            else:
                share_mpi([result])
                return result
        else:
            result = share_mpi()
            if result is None:
                raise OtherProcessError('Root errored in %s' % func.__name__)
            return result[0]

    return wrapper


def set_from_root(attributes):
    atts = [attributes] if isinstance(attributes, str) else attributes

    def set_method(method):

        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            if is_main_process():
                try:
                    result = method(self, *args, **kwargs)
                except Exception:
                    share_mpi()
                    raise
                else:
                    share_mpi([result] + [getattr(self, var, None) for var in atts])
            else:
                values = share_mpi()
                if values is None:
                    raise OtherProcessError('Root errored in %s' % method.__name__)
                for name, var in zip(atts, values[1:]):
                    setattr(self, name, var)
                result = values[0]
            return result

        return wrapper

    return set_method


def sync_errors(func):
    err = 'Another process raised an error in %s' % func.__name__

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except Exception:
            allgather(True)
            raise
        else:
            if any(allgather(False)):
                raise OtherProcessError(err)
            return result

    return wrapper


# Wrapper for main functions. Traps MPI deadlock via timeout MPI_ABORT if needed,

def sync_state(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not more_than_one_process():
            return func(*args, **kwargs)

        with ProcessState(func.__name__):
            return func(*args, **kwargs)

    return wrapper


# Synchronization between processes, generating OtherProcessError in other threads if
# any one process raises an exception

class State(IntEnum):
    NONE = 0
    READY = 1
    END = 2
    ERROR = 3


class SyncError(OtherProcessError):
    pass


# log = logging.getLogger('state')
log: Any = None


class ProcessState:
    _id: int = 1

    def __init__(self, name='error',
                 time_out_seconds: Union[float, int] = default_error_timeout_seconds,
                 sleep_interval=0.01,
                 timeout_abort_proc: Callable = abort_if_mpi):
        self.name = str(name)
        ProcessState._id += 1
        self.tag = ProcessState._id
        self.states = np.empty(size(), dtype=int)
        self.sleep_interval = sleep_interval
        self.time_out_seconds = time_out_seconds
        self._recv_state = np.empty(1, dtype=int)
        self._rank = rank()
        self._others = [i for i in range(size()) if i != self._rank]
        self.timeout_abort_proc = timeout_abort_proc

    def set(self, value: State):
        """
        Sends an error signal to the other MPI processes.
        """
        if self.states[self._rank] != value:
            if log:
                log.info('SET %s %s %s', self.name, self.tag, value)
            self.states[self._rank] = value
            for i_rank in self._others:
                _mpi_comm.Isend(self.states[self._rank],
                                dest=i_rank, tag=self.tag).Test()
            return True
        else:
            return False

    @more_than_one
    def sync(self, check_error=False):
        """
        Gets any messages from other processes without waiting, and optionally
        raises error if any others are in error state.
        """
        while _mpi_comm.iprobe(source=_mpi.ANY_SOURCE, tag=self.tag):
            status = _mpi.Status()
            _mpi_comm.Recv(self._recv_state, source=_mpi.ANY_SOURCE, tag=self.tag,
                           status=status)
            state = self._recv_state[0]
            self.states[status.Get_source()] = state
            if check_error and state == State.ERROR:
                self.fire_error(SyncError)
            if log:
                log.info('SYNC %s %s %s', self.name, self.tag, self.states)

    def check_error(self):
        """
        Raises error if any other processes in error state
        """
        self.sync(check_error=True)

    def fire_error(self, cls=OtherProcessError, msg=_other_process_msg):
        raise cls("[%s: %s] %s" % (rank(), self.name, msg))

    def wait_all_ended(self, timeout=False):
        """
        Wait until all processes in ERROR or END state.
        """
        self.sync()
        time_start = time.time()
        while any(self.states < State.END):
            time.sleep(self.sleep_interval)
            if timeout and time.time() - time_start > self.time_out_seconds:
                return False
            self.sync()
        return True

    def all_ready(self) -> bool:
        """
        Test is all processes in READY state (and if they are reset to NONE).
        """
        self.sync(check_error=True)
        all_ready = all(self.states == State.READY)
        if all_ready:
            self.states[:] = State.NONE
        return all_ready

    def __enter__(self):
        if more_than_one_process():
            self.tag = share(self.tag)
            self.last_process_state = process_state
            set_current_process_state(self)
        self.states[:] = State.NONE
        return self

    @more_than_one
    def __exit__(self, exc_type, exc_val, exc_tb):
        if log:
            log.info('END %s %s', self.name, self.tag)
        if exc_type:
            self.set(State.ERROR)
            if not self.wait_all_ended(
                    timeout=not issubclass(exc_type, OtherProcessError)):
                from cobaya.log import get_traceback_text, LoggedError, get_logger
                get_logger(self.name).critical(
                    "Aborting MPI due to error" if issubclass(exc_type, LoggedError) else
                    get_traceback_text(sys.exc_info()))
                self.timeout_abort_proc()
                self.wait_all_ended()  # if didn't actually MPI abort
        else:
            self.set(State.END)
            self.wait_all_ended()
        set_current_process_state(self.last_process_state)
        if not exc_type and any(self.states == State.ERROR):
            self.fire_error()


process_state: Optional[ProcessState] = None


def set_current_process_state(state: Optional[ProcessState]):
    global process_state
    process_state = state



from cobaya.likelihoods.base_classes import BAO


class sdss_dr12_consensus_final(BAO):
    r"""
    Likelihood of the combined power spectrum BAO and full-shape measurements of
    SDSS DR12 \cite{Alam:2016hwk}.
    """
    pass
